import math
from datetime import timedelta, datetime
from typing import Dict, Any, Optional

import croniter
from django.conf import settings as django_settings
from django.contrib import admin
from django.contrib.contenttypes.fields import GenericRelation
from django.core.exceptions import ValidationError
from django.core.mail import mail_admins
from django.db import models
from django.templatetags.tz import utc
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from scheduler import settings
from scheduler.helpers.callback import Callback
from scheduler.helpers.queues import Queue
from scheduler.helpers.queues import get_queue
from scheduler.redis_models import JobModel
from scheduler.settings import logger, get_queue_names
from scheduler.types import ConnectionType, TASK_TYPES
from .args import TaskArg, TaskKwarg
from ..helpers import utils


def _get_task_for_job(job: JobModel) -> Optional["Task"]:
    if job.task_type is None or job.scheduled_task_id is None:
        return None
    task = Task.objects.filter(id=job.scheduled_task_id).first()
    return task


def failure_callback(job: JobModel, connection, result, *args, **kwargs):
    task = _get_task_for_job(job)
    if task is None:
        logger.warn(f"Could not find task for job {job.name}")
        return
    mail_admins(
        f"Task {task.id}/{task.name} has failed",
        "See django-admin for logs",
    )
    task.job_name = None
    task.failed_runs += 1
    task.last_failed_run = timezone.now()
    task.save(schedule_job=True)


def success_callback(job: JobModel, connection: ConnectionType, result: Any, *args, **kwargs):
    task = _get_task_for_job(job)
    if task is None:
        logger.warn(f"Could not find task for job {job.name}")
        return
    task.job_name = None
    task.successful_runs += 1
    task.last_successful_run = timezone.now()
    task.save(schedule_job=True)


def get_queue_choices():
    queue_names = get_queue_names()
    return [(queue, queue) for queue in queue_names]


class TaskType(models.TextChoices):
    CRON = "CronTaskType", _("Cron Task")
    REPEATABLE = "RepeatableTaskType", _("Repeatable Task")
    ONCE = "OnceTaskType", _("Run once")


class Task(models.Model):
    class TimeUnits(models.TextChoices):
        SECONDS = "seconds", _("seconds")
        MINUTES = "minutes", _("minutes")
        HOURS = "hours", _("hours")
        DAYS = "days", _("days")
        WEEKS = "weeks", _("weeks")

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    name = models.CharField(_("name"), max_length=128, unique=True, help_text=_("Name of the job"))
    task_type = models.CharField(_("Task type"), max_length=32, choices=TaskType.choices, default=TaskType.ONCE)
    callable = models.CharField(_("callable"), max_length=2048)
    callable_args = GenericRelation(TaskArg, related_query_name="args")
    callable_kwargs = GenericRelation(TaskKwarg, related_query_name="kwargs")
    enabled = models.BooleanField(
        _("enabled"),
        default=True,
        help_text=_(
            "Should job be scheduled? This field is useful to keep past jobs that should no longer be scheduled"
        ),
    )
    queue = models.CharField(_("queue"), max_length=255, choices=get_queue_choices, help_text=_("Queue name"))
    job_name = models.CharField(
        _("job name"), max_length=128, editable=False, blank=True, null=True, help_text=_("Current job_name on queue")
    )
    at_front = models.BooleanField(
        _("At front"),
        default=False,
        help_text=_("When queuing the job, add it in the front of the queue"),
    )
    timeout = models.IntegerField(
        _("timeout"),
        blank=True,
        null=True,
        help_text=_(
            "Timeout specifies the maximum runtime, in seconds, for the job "
            "before it'll be considered 'lost'. Blank uses the default "
            "timeout."
        ),
    )
    result_ttl = models.IntegerField(
        _("result ttl"),
        blank=True,
        null=True,
        help_text=mark_safe(
            """The TTL value (in seconds) of the job result.<br/>
               -1: Result never expires, you should delete jobs manually. <br/>
                0: Result gets deleted immediately. <br/>
                >0: Result expires after n seconds."""
        ),
    )
    failed_runs = models.PositiveIntegerField(
        _("failed runs"),
        default=0,
        help_text=_("Number of times the task has failed"),
    )
    successful_runs = models.PositiveIntegerField(
        _("successful runs"),
        default=0,
        help_text=_("Number of times the task has succeeded"),
    )
    last_successful_run = models.DateTimeField(
        _("last successful run"),
        blank=True,
        null=True,
        help_text=_("Last time the task has succeeded"),
    )
    last_failed_run = models.DateTimeField(
        _("last failed run"),
        blank=True,
        null=True,
        help_text=_("Last time the task has failed"),
    )
    interval = models.PositiveIntegerField(
        _("interval"),
        blank=True,
        null=True,
        help_text=_("Interval for repeatable task"),
    )
    interval_unit = models.CharField(
        _("interval unit"),
        max_length=12,
        choices=TimeUnits.choices,
        default=TimeUnits.HOURS,
        blank=True,
        null=True,
    )
    repeat = models.PositiveIntegerField(
        _("repeat"),
        blank=True,
        null=True,
        help_text=_("Number of times to run the job. Leaving this blank means it will run forever."),
    )
    scheduled_time = models.DateTimeField(_("scheduled time"), blank=True, null=True)
    cron_string = models.CharField(
        _("cron string"),
        max_length=64,
        blank=True,
        null=True,
        help_text=mark_safe(
            """Define the schedule in a crontab like syntax.
            Times are in UTC. Use <a href="https://crontab.guru/">crontab.guru</a> to create a cron string."""
        ),
    )

    def callable_func(self):
        """Translate callable string to callable"""
        return utils.callable_func(self.callable)

    @admin.display(boolean=True, description=_("is scheduled?"))
    def is_scheduled(self) -> bool:
        """Check whether a next job for this task is queued/scheduled to be executed"""
        if self.job_name is None:  # no job_id => is not scheduled
            return False
        # check whether job_id is in scheduled/queued/active jobs
        res = (
            (self.job_name in self.rqueue.scheduled_job_registry.all())
            or (self.job_name in self.rqueue.queued_job_registry.all())
            or (self.job_name in self.rqueue.active_job_registry.all())
        )
        # If the job_id is not scheduled/queued/started,
        # update the job_id to None. (The job_id belongs to a previous run which is completed)
        if not res:
            self.job_name = None
            super(Task, self).save()
        return res

    @admin.display(description="Callable")
    def function_string(self) -> str:
        args = self.parse_args()
        args_list = [repr(arg) for arg in args]
        kwargs = self.parse_kwargs()
        kwargs_list = [k + "=" + repr(v) for (k, v) in kwargs.items()]
        return self.callable + f"({', '.join(args_list + kwargs_list)})"

    def parse_args(self):
        """Parse args for running the job"""
        args = self.callable_args.all()
        return [arg.value() for arg in args]

    def parse_kwargs(self):
        """Parse kwargs for running the job"""
        kwargs = self.callable_kwargs.all()
        return dict([kwarg.value() for kwarg in kwargs])

    def _next_job_id(self):
        addition = timezone.now().strftime("%Y%m%d%H%M%S%f")
        return f"{self.queue}:{self.id}:{addition}"

    def _enqueue_args(self) -> Dict:
        """Args for Queue.enqueue_call.
        Set all arguments for Queue.enqueue. Particularly:
        - set job timeout and ttl
        - ensure a callback to reschedule the job next iteration.
        - Set job-id to proper format
        - set job meta
        """
        res = dict(
            meta=dict(),
            task_type=self.task_type,
            scheduled_task_id=self.id,
            on_success=Callback(success_callback),
            on_failure=Callback(failure_callback),
            name=self._next_job_id(),
        )
        if self.at_front:
            res["at_front"] = self.at_front
        if self.timeout:
            res["timeout"] = self.timeout
        if self.result_ttl is not None:
            res["result_ttl"] = self.result_ttl
        if self.task_type == TaskType.REPEATABLE:
            res["meta"]["interval"] = self.interval_seconds()
            res["meta"]["repeat"] = self.repeat
        return res

    @property
    def rqueue(self) -> Queue:
        """Returns django-queue for job"""
        return get_queue(self.queue)

    def enqueue_to_run(self) -> bool:
        """Enqueue task to run now as a different instance from the scheduled task."""
        kwargs = self._enqueue_args()
        self.rqueue.create_and_enqueue_job(run_task, args=(self.task_type, self.id), when=None, **kwargs)
        return True

    def unschedule(self) -> bool:
        """Remove a job from django-queue.

        If a job is queued to be executed or scheduled to be executed, it will remove it.
        """
        if self.job_name is not None:
            self.rqueue.delete_job(self.job_name)
            self.job_name = None
        self.save(schedule_job=False)
        return True

    def _schedule_time(self) -> datetime:
        if self.task_type == TaskType.CRON:
            self.scheduled_time = get_next_cron_time(self.cron_string)
        elif self.task_type == TaskType.REPEATABLE:
            _now = timezone.now()
            if self.scheduled_time >= _now:
                return utc(self.scheduled_time) if django_settings.USE_TZ else self.scheduled_time
            gap = math.ceil((_now.timestamp() - self.scheduled_time.timestamp()) / self.interval_seconds())
            if self.repeat is None or self.repeat >= gap:
                self.scheduled_time += timedelta(seconds=self.interval_seconds() * gap)
                self.repeat = (self.repeat - gap) if self.repeat is not None else None
        return utc(self.scheduled_time) if django_settings.USE_TZ else self.scheduled_time

    def to_dict(self) -> Dict:
        """Export model to dictionary, so it can be saved as external file backup"""
        interval_unit = str(self.interval_unit) if self.interval_unit else None
        res = dict(
            model=str(self.task_type),
            name=self.name,
            callable=self.callable,
            callable_args=[
                dict(
                    arg_type=arg.arg_type,
                    val=arg.val,
                )
                for arg in self.callable_args.all()
            ],
            callable_kwargs=[
                dict(
                    arg_type=arg.arg_type,
                    key=arg.key,
                    val=arg.val,
                )
                for arg in self.callable_kwargs.all()
            ],
            enabled=self.enabled,
            queue=self.queue,
            repeat=getattr(self, "repeat", None),
            at_front=self.at_front,
            timeout=self.timeout,
            result_ttl=self.result_ttl,
            cron_string=getattr(self, "cron_string", None),
            scheduled_time=self._schedule_time().isoformat(),
            interval=getattr(self, "interval", None),
            interval_unit=interval_unit,
            successful_runs=getattr(self, "successful_runs", None),
            failed_runs=getattr(self, "failed_runs", None),
            last_successful_run=getattr(self, "last_successful_run", None),
            last_failed_run=getattr(self, "last_failed_run", None),
        )
        return res

    def get_absolute_url(self):
        model = self._meta.model.__name__.lower()
        return reverse(
            f"admin:scheduler_{model}_change",
            args=[
                self.id,
            ],
        )

    def __str__(self):
        func = self.function_string()
        return f"{self.task_type}[{self.name}={func}]"

    def _schedule(self) -> bool:
        """Schedule the next execution for the task to run.
        :returns: True if a job was scheduled, False otherwise.
        """
        self.refresh_from_db()
        if self.is_scheduled():
            logger.debug(f"Task {self.name} already scheduled")
            return False
        if not self.enabled:
            logger.debug(f"Task {str(self)} disabled, enable task before scheduling")
            return False
        schedule_time = self._schedule_time()
        if self.task_type in {TaskType.REPEATABLE, TaskType.ONCE} and schedule_time < timezone.now():
            logger.debug(f"Task {str(self)} scheduled time is in the past, not scheduling")
            return False
        kwargs = self._enqueue_args()
        job = self.rqueue.create_and_enqueue_job(
            run_task,
            args=(self.task_type, self.id),
            when=schedule_time,
            **kwargs,
        )
        self.job_name = job.name
        return True

    def save(self, **kwargs):
        schedule_job = kwargs.pop("schedule_job", True)
        update_fields = kwargs.get("update_fields", None)
        if update_fields is not None:
            kwargs["update_fields"] = set(update_fields).union({"updated_at"})
        super(Task, self).save(**kwargs)
        if schedule_job:
            self._schedule()
            super(Task, self).save()

    def delete(self, **kwargs):
        self.unschedule()
        super(Task, self).delete(**kwargs)

    def interval_seconds(self):
        kwargs = {
            self.interval_unit: self.interval,
        }
        return timedelta(**kwargs).total_seconds()

    def clean_callable(self):
        try:
            utils.callable_func(self.callable)
        except Exception:
            raise ValidationError(
                {"callable": ValidationError(_("Invalid callable, must be importable"), code="invalid")}
            )

    def clean_queue(self):
        queue_names = get_queue_names()
        if self.queue not in queue_names:
            raise ValidationError(
                {
                    "queue": ValidationError(
                        "Invalid queue, must be one of: {}".format(", ".join(queue_names)), code="invalid"
                    )
                }
            )

    def clean_interval_unit(self):
        config = settings.SCHEDULER_CONFIG
        if config.SCHEDULER_INTERVAL > self.interval_seconds():
            raise ValidationError(
                _("Job interval is set lower than %(queue)r queue's interval. minimum interval is %(interval)"),
                code="invalid",
                params={"queue": self.queue, "interval": config.SCHEDULER_INTERVAL},
            )
        if self.interval_seconds() % config.SCHEDULER_INTERVAL:
            raise ValidationError(
                _("Job interval is not a multiple of rq_scheduler's interval frequency: %(interval)ss"),
                code="invalid",
                params={"interval": config.SCHEDULER_INTERVAL},
            )

    def clean_result_ttl(self) -> None:
        """Throws an error if there are repeats left to run and the result_ttl won't last until the next scheduled time.
        :return: None
        """
        if self.result_ttl and self.result_ttl != -1 and self.result_ttl < self.interval_seconds() and self.repeat:
            raise ValidationError(
                _(
                    "Job result_ttl must be either indefinite (-1) or "
                    "longer than the interval, %(interval)s seconds, to ensure rescheduling."
                ),
                code="invalid",
                params={"interval": self.interval_seconds()},
            )

    def clean_cron_string(self):
        try:
            croniter.croniter(self.cron_string)
        except ValueError as e:
            raise ValidationError({"cron_string": ValidationError(_(str(e)), code="invalid")})

    def clean(self):
        self.clean_queue()
        self.clean_callable()
        if self.task_type == TaskType.CRON:
            self.clean_cron_string()
        if self.task_type == TaskType.REPEATABLE:
            self.clean_interval_unit()
            self.clean_result_ttl()
        if self.task_type == TaskType.ONCE and self.scheduled_time is None:
            raise ValidationError({"scheduled_time": ValidationError(_("Scheduled time is required"), code="invalid")})
        if self.task_type == TaskType.ONCE and self.scheduled_time < timezone.now():
            raise ValidationError(
                {"scheduled_time": ValidationError(_("Scheduled time must be in the future"), code="invalid")}
            )


def get_next_cron_time(cron_string: Optional[str]) -> Optional[timezone.datetime]:
    """Calculate the next scheduled time by creating a crontab object with a cron string"""
    if cron_string is None:
        return None
    now = timezone.now()
    itr = croniter.croniter(cron_string, now)
    next_itr = itr.get_next(timezone.datetime)
    return next_itr


def get_scheduled_task(task_type_str: str, task_id: int) -> Task:
    # Try with new model names
    if task_type_str in TASK_TYPES:
        try:
            task_type = TaskType(task_type_str)
            task = Task.objects.filter(task_type=task_type, id=task_id).first()
            if task is None:
                raise ValueError(f"Job {task_type}:{task_id} does not exit")
            return task
        except ValueError:
            raise ValueError(f"Invalid task type {task_type_str}")
    raise ValueError(f"Job Model {task_type_str} does not exist, choices are {TASK_TYPES}")


def run_task(task_model: str, task_id: int) -> Any:
    """Run a scheduled job"""
    if isinstance(task_id, str):
        task_id = int(task_id)
    scheduled_task = get_scheduled_task(task_model, task_id)
    logger.debug(f"Running task {str(scheduled_task)}")
    args = scheduled_task.parse_args()
    kwargs = scheduled_task.parse_kwargs()
    res = scheduled_task.callable_func()(*args, **kwargs)
    return res



from django.urls import reverse

from scheduler.helpers.queues import get_queue
from scheduler.redis_models import JobStatus, JobModel
from scheduler.tests.jobs import failing_job, test_job
from scheduler.tests.test_views.base import BaseTestCase
from scheduler.tests.testtools import assert_message_in_response
from scheduler.worker import create_worker


class QueueActionsViewsTest(BaseTestCase):
    def test_job_list_action_delete_jobs__with_bad_next_url(self):
        queue = get_queue("django_tasks_scheduler_test")

        # enqueue some jobs
        job_names = []
        for _ in range(0, 3):
            job = queue.create_and_enqueue_job(test_job, job_info_ttl=0)
            job_names.append(job.name)

        # remove those jobs using view
        res = self.client.post(
            reverse("queue_job_actions", args=[queue.name]),
            {
                "action": "delete",
                "job_names": job_names,
                "next_url": "bad_url",
            },
            follow=True,
        )
        assert_message_in_response(res, "Bad followup URL")
        # check if jobs are removed
        self.assertEqual(200, res.status_code)
        for job_name in job_names:
            self.assertFalse(JobModel.exists(job_name, connection=queue.connection), f"job {job_name} exists")
            self.assertNotIn(job_name, queue.queued_job_registry.all())

    def test_job_list_action_delete_jobs(self):
        queue = get_queue("django_tasks_scheduler_test")

        # enqueue some jobs
        job_names = []
        for _ in range(0, 3):
            job = queue.create_and_enqueue_job(test_job, job_info_ttl=0)
            job_names.append(job.name)

        # remove those jobs using view
        res = self.client.post(
            reverse("queue_job_actions", args=[queue.name]),
            {"action": "delete", "job_names": job_names},
            follow=True,
        )

        # check if jobs are removed
        self.assertEqual(200, res.status_code)
        for job_name in job_names:
            self.assertFalse(JobModel.exists(job_name, connection=queue.connection), f"job {job_name} exists")
            self.assertNotIn(job_name, queue.queued_job_registry.all())

    def test_job_list_action_requeue_jobs(self):
        queue_name = "django_tasks_scheduler_test"
        queue = get_queue(queue_name)

        # enqueue some jobs that will fail
        job_names = []
        for _ in range(0, 3):
            job = queue.create_and_enqueue_job(failing_job)
            job_names.append(job.name)

        # do those jobs = fail them
        worker = create_worker(queue_name, burst=True)
        worker.work()

        # check if all jobs are really failed
        for job_name in job_names:
            job = JobModel.get(job_name, connection=queue.connection)
            self.assertTrue(job.is_failed)

        # re-nqueue failed jobs from failed queue
        self.client.post(reverse("queue_job_actions", args=[queue_name]), {"action": "requeue", "job_names": job_names})

        # check if we requeue all failed jobs
        for job_name in job_names:
            job = JobModel.get(job_name, connection=queue.connection)
            self.assertFalse(job.is_failed)

    def test_job_list_action_stop_jobs__move_to_finished_registry(self):
        queue_name = "django_tasks_scheduler_test"
        queue = get_queue(queue_name)

        # Enqueue some jobs
        job_names = []
        worker = create_worker(queue_name)
        worker.bootstrap()
        for _ in range(3):
            job = queue.create_and_enqueue_job(test_job)
            job_names.append(job.name)
            worker.worker_before_execution(job, connection=queue.connection)
            job.prepare_for_execution(worker.name, queue.active_job_registry, connection=queue.connection)

        # Check if the jobs are started
        for job_name in job_names:
            job = JobModel.get(job_name, connection=queue.connection)
            self.assertEqual(job.status, JobStatus.STARTED)

        # Stop those jobs using the view
        self.assertEqual(len(queue.active_job_registry), len(job_names))
        self.client.post(reverse("queue_job_actions", args=[queue_name]), {"action": "stop", "job_names": job_names})
        self.assertEqual(0, len(queue.active_job_registry))

        self.assertEqual(0, len(queue.canceled_job_registry))
        self.assertEqual(len(job_names), len(queue.finished_job_registry))

        for job_name in job_names:
            self.assertIn(job_name, queue.finished_job_registry)



import dataclasses
import json
from collections.abc import Sequence
from datetime import datetime, timezone
from enum import Enum
from typing import List, Optional, Union, Dict, Collection, Any, ClassVar, Set, Type

from redis import Redis

from scheduler.settings import logger
from scheduler.types import ConnectionType, Self

MAX_KEYS = 1000


def as_str(v: Union[bytes, str]) -> Optional[str]:
    """Converts a `bytes` value to a string using `utf-8`.

    :param v: The value (None/bytes/str)
    :raises: ValueError: If the value is not `bytes` or `str`
    :returns: Either the decoded string or None
    """
    if v is None or isinstance(v, str):
        return v
    if isinstance(v, bytes):
        return v.decode("utf-8")
    raise ValueError(f"Unknown type {type(v)} for `{v}`.")


def decode_dict(d: Dict[bytes, bytes], exclude_keys: Set[str]) -> Dict[str, str]:
    return {k.decode(): v.decode() for (k, v) in d.items() if k.decode() not in exclude_keys}


def _serialize(value: Any) -> Optional[Any]:
    if value is None:
        return None
    if isinstance(value, bool):
        value = int(value)
    elif isinstance(value, Enum):
        value = value.value
    elif isinstance(value, datetime):
        value = value.isoformat()
    elif isinstance(value, dict):
        value = json.dumps(value)
    elif isinstance(value, (int, float)):
        return value
    elif isinstance(value, (list, set, tuple)):
        return json.dumps(value, default=str)
    return str(value)


def _deserialize(value: str, _type: Type) -> Any:
    if value is None:
        return None
    try:
        if _type is str or _type == Optional[str]:
            return as_str(value)
        if _type is datetime or _type == Optional[datetime]:
            return datetime.fromisoformat(as_str(value))
        elif _type is bool:
            return bool(int(value))
        elif _type is int or _type == Optional[int]:
            return int(value)
        elif _type is float or _type == Optional[float]:
            return float(value)
        elif _type in {List[str], Dict[str, str]}:
            return json.loads(value)
        elif _type == Optional[Any]:
            return json.loads(value)
        elif issubclass(_type, Enum):
            return _type(as_str(value))
    except (ValueError, TypeError) as e:
        logger.warning(f"Failed to deserialize {value} as {_type}: {e}")
    return value


@dataclasses.dataclass(slots=True, kw_only=True)
class BaseModel:
    name: str
    _element_key_template: ClassVar[str] = ":element:{}"
    # fields that are not serializable using method above and should be dealt with in the subclass
    # e.g. args/kwargs for a job
    _non_serializable_fields: ClassVar[Set[str]] = set()

    @classmethod
    def key_for(cls, name: str) -> str:
        return cls._element_key_template.format(name)

    @property
    def _key(self) -> str:
        return self._element_key_template.format(self.name)

    def serialize(self, with_nones: bool = False) -> Dict[str, str]:
        data = dataclasses.asdict(
            self, dict_factory=lambda fields: {key: value for (key, value) in fields if not key.startswith("_")}
        )
        if not with_nones:
            data = {k: v for k, v in data.items() if v is not None and k not in self._non_serializable_fields}
        for k in data:
            data[k] = _serialize(data[k])
        return data

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> Self:
        types = {f.name: f.type for f in dataclasses.fields(cls) if f.name not in cls._non_serializable_fields}
        for k in data:
            if k not in types:
                logger.warning(f"Unknown field {k} in {cls.__name__}")
                continue
            data[k] = _deserialize(data[k], types[k])
        return cls(**data)


@dataclasses.dataclass(slots=True, kw_only=True)
class HashModel(BaseModel):
    created_at: Optional[datetime] = None
    parent: Optional[str] = None
    _dirty_fields: Set[str] = dataclasses.field(default_factory=set)  # fields that were changed
    _save_all: bool = True  # Save all fields to broker, after init, or after delete
    _list_key: ClassVar[str] = ":list_all:"
    _children_key_template: ClassVar[str] = ":children:{}:"

    def __post_init__(self):
        self._dirty_fields = set()
        self._save_all = True

    def __setattr__(self, key, value):
        if key != "_dirty_fields" and hasattr(self, "_dirty_fields"):
            self._dirty_fields.add(key)
        super(HashModel, self).__setattr__(key, value)

    @property
    def _parent_key(self) -> Optional[str]:
        if self.parent is None:
            return None
        return self._children_key_template.format(self.parent)

    @classmethod
    def all_names(cls, connection: Redis, parent: Optional[str] = None) -> Collection[str]:
        collection_key = cls._children_key_template.format(parent) if parent else cls._list_key
        collection_members = connection.smembers(collection_key)
        return [r.decode() for r in collection_members]

    @classmethod
    def all(cls, connection: Redis, parent: Optional[str] = None) -> List[Self]:
        keys = cls.all_names(connection, parent)
        items = [cls.get(k, connection) for k in keys]
        return [w for w in items if w is not None]

    @classmethod
    def exists(cls, name: str, connection: ConnectionType) -> bool:
        if name is None:
            return False
        return connection.exists(cls._element_key_template.format(name)) > 0

    @classmethod
    def delete_many(cls, names: List[str], connection: ConnectionType) -> None:
        for name in names:
            connection.delete(cls._element_key_template.format(name))

    @classmethod
    def get(cls, name: str, connection: ConnectionType) -> Optional[Self]:
        res = connection.hgetall(cls._element_key_template.format(name))
        if not res:
            return None
        try:
            return cls.deserialize(decode_dict(res, set()))
        except Exception as e:
            logger.warning(f"Failed to deserialize {name}: {e}")
            return None

    @classmethod
    def get_many(cls, names: Sequence[str], connection: ConnectionType) -> List[Self]:
        pipeline = connection.pipeline()
        for name in names:
            pipeline.hgetall(cls._element_key_template.format(name))
        values = pipeline.execute()
        return [(cls.deserialize(decode_dict(v, set())) if v else None) for v in values]

    def save(self, connection: ConnectionType) -> None:
        connection.sadd(self._list_key, self.name)
        if self._parent_key is not None:
            connection.sadd(self._parent_key, self.name)
        mapping = self.serialize(with_nones=True)
        if not self._save_all and len(self._dirty_fields) > 0:
            mapping = {k: v for k, v in mapping.items() if k in self._dirty_fields}
        none_values = {k for k, v in mapping.items() if v is None}
        if none_values:
            connection.hdel(self._key, *none_values)
        mapping = {k: v for k, v in mapping.items() if v is not None}
        if mapping:
            connection.hset(self._key, mapping=mapping)
        self._dirty_fields = set()
        self._save_all = False

    def delete(self, connection: ConnectionType) -> None:
        connection.srem(self._list_key, self._key)
        if self._parent_key is not None:
            connection.srem(self._parent_key, 0, self._key)
        connection.delete(self._key)
        self._save_all = True

    @classmethod
    def count(cls, connection: ConnectionType, parent: Optional[str] = None) -> int:
        if parent is not None:
            result = connection.scard(cls._children_key_template.format(parent))
        else:
            result = connection.scard(cls._list_key)
        return result

    def get_field(self, field: str, connection: ConnectionType) -> Any:
        types = {f.name: f.type for f in dataclasses.fields(self)}
        res = connection.hget(self._key, field)
        return _deserialize(res, types[field])

    def set_field(self, field: str, value: Any, connection: ConnectionType, set_attribute: bool = True) -> None:
        if not hasattr(self, field):
            raise AttributeError(f"Field {field} does not exist")
        if set_attribute:
            setattr(self, field, value)
        if value is None:
            connection.hdel(self._key, field)
            return
        value = _serialize(value)
        connection.hset(self._key, field, value)


@dataclasses.dataclass(slots=True, kw_only=True)
class StreamModel(BaseModel):
    _children_key_template: ClassVar[str] = ":children:{}:"

    def __init__(self, name: str, parent: str, created_at: Optional[datetime] = None):
        self.name = name
        self.created_at: datetime = created_at or datetime.now(timezone.utc)
        self.parent: str = parent

    @property
    def _parent_key(self) -> str:
        return self._children_key_template.format(self.parent)

    @classmethod
    def all(cls, connection: ConnectionType, parent: str) -> List[Self]:
        results = connection.xrevrange(cls._children_key_template.format(parent), "+", "-")
        return [cls.deserialize(decode_dict(result[1], exclude_keys=set())) for result in results]

    def save(self, connection: ConnectionType) -> bool:
        result = connection.xadd(self._parent_key, self.serialize(), maxlen=10)
        return bool(result)



import datetime
import importlib
import time
from typing import Callable


def current_timestamp() -> int:
    """Returns current UTC timestamp in secs"""
    return int(time.time())


def utcnow() -> datetime.datetime:
    """Return now in UTC"""
    return datetime.datetime.now(datetime.timezone.utc)


def callable_func(callable_str: str) -> Callable:
    path = callable_str.split(".")
    module = importlib.import_module(".".join(path[:-1]))
    func = getattr(module, path[-1])
    if callable(func) is False:
        raise TypeError(f"'{callable_str}' is not callable")
    return func



from django.urls import path

from . import views

urlpatterns = [
    path("queues/", views.stats, name="queues_home"),
    path("queues/stats.json", views.stats_json, name="queues_home_json"),
    path("queues/<str:queue_name>/workers/", views.queue_workers, name="queue_workers"),
    path("queues/<str:queue_name>/<str:registry_name>/jobs", views.list_registry_jobs, name="queue_registry_jobs"),
    path(
        "queues/<str:queue_name>/<str:registry_name>/<str:action>/",
        views.queue_registry_actions,
        name="queue_registry_action",
    ),
    path("queues/<str:queue_name>/confirm-action/", views.queue_confirm_job_action, name="queue_confirm_job_action"),
    path("queues/<str:queue_name>/actions/", views.queue_job_actions, name="queue_job_actions"),
]

urlpatterns += [
    path("workers/", views.workers_list, name="workers_home"),
    path("workers/<str:name>/", views.worker_details, name="worker_details"),
    path("jobs/<str:job_name>/", views.job_detail, name="job_details"),
    path("jobs/<str:job_name>/<str:action>/", views.job_action, name="job_detail_action"),
]



import os
import signal
from typing import Dict, Any

from scheduler.types import ConnectionType
from scheduler.redis_models import WorkerModel, JobModel
from scheduler.settings import logger
from scheduler.worker.commands.worker_commands import WorkerCommand, WorkerCommandError


class StopJobCommand(WorkerCommand):
    """stop-job command"""

    command_name = "stop-job"

    def __init__(self, *args, job_name: str, worker_name: str, **kwargs) -> None:
        super().__init__(*args, worker_name=worker_name, **kwargs)
        self.job_name = job_name
        if self.job_name is None:
            raise WorkerCommandError("job_name for kill-job command is required")

    def command_payload(self) -> Dict[str, Any]:
        return super().command_payload(job_name=self.job_name)

    def process_command(self, connection: ConnectionType) -> None:
        logger.debug(f"Received command to stop job {self.job_name}")
        worker_model = WorkerModel.get(self.worker_name, connection)
        job_model = JobModel.get(self.job_name, connection)
        if worker_model is None:
            logger.error(f"Worker {self.worker_name} not found")
            return
        if job_model is None:
            logger.error(f"Job {self.job_name} not found")
            return
        if worker_model.pid == worker_model.job_execution_process_pid:
            logger.warning(f"Job execution process ID and worker process id {worker_model.pid} are equal, skipping")
            return
        if not worker_model.job_execution_process_pid:
            logger.error(f"Worker {self.worker_name} has no job execution process")
            return
        if worker_model.current_job_name != self.job_name:
            logger.info(
                f"{self.worker_name} working on job {worker_model.current_job_name}, "
                f"not on {self.job_name}, kill-job command ignored."
            )
            return
        worker_model.set_field("stopped_job_name", self.job_name, connection)
        try:
            pgid = os.getpgid(worker_model.job_execution_process_pid)
            logger.debug(
                f"worker_pid {worker_model.pid}, job_execution_process {worker_model.job_execution_process_pid}"
            )
            if pgid == worker_model.pid:
                logger.error("No separate process for job execution, skipping")
                return
            os.killpg(pgid, signal.SIGTERM)
        except ProcessLookupError as e:
            logger.error(f"Error killing job {self.job_name}: {e}")



from functools import wraps
from typing import Any, Callable, Dict, Optional, Union, List

from scheduler.helpers.callback import Callback
from scheduler.types import ConnectionType

JOB_METHODS_LIST: List[str] = list()


class job:
    def __init__(
        self,
        queue: Union["Queue", str, None] = None,  # noqa: F821
        connection: Optional[ConnectionType] = None,
        timeout: Optional[int] = None,
        result_ttl: Optional[int] = None,
        job_info_ttl: Optional[int] = None,
        at_front: bool = False,
        meta: Optional[Dict[Any, Any]] = None,
        description: Optional[str] = None,
        on_failure: Optional[Union["Callback", Callable[..., Any]]] = None,
        on_success: Optional[Union["Callback", Callable[..., Any]]] = None,
        on_stopped: Optional[Union["Callback", Callable[..., Any]]] = None,
    ):
        """A decorator that adds a ``delay`` method to the decorated function, which in turn creates a RQ job when
        called. Accepts a required ``queue`` argument that can be either a ``Queue`` instance or a string
        denoting the queue name.  For example::


        >>> @job(queue='default')
        >>> def simple_add(x, y):
        >>>    return x + y
        >>> ...
        >>> # Puts `simple_add` function into queue
        >>> simple_add.delay(1, 2)

        :param queue: The queue to use, can be the Queue class itself, or the queue name (str)
        :type queue: Union['Queue', str]
        :param connection: Broker Connection
        :param timeout: Job timeout
        :param result_ttl: Result time to live
        :param job_info_ttl: Time to live for job info
        :param at_front: Whether to enqueue the job at front of the queue
        :param meta: Arbitraty metadata about the job
        :param description: Job description
        :param on_failure: Callable to run on failure
        :param on_success: Callable to run on success
        :param on_stopped: Callable to run when stopped
        """
        from scheduler.helpers.queues import get_queue

        if queue is None:
            queue = "default"
        self.queue = get_queue(queue) if isinstance(queue, str) else queue
        self.connection = connection
        self.timeout = timeout
        self.result_ttl = result_ttl
        self.job_info_ttl = job_info_ttl
        self.meta = meta
        self.at_front = at_front
        self.description = description
        self.on_success = on_success
        self.on_failure = on_failure
        self.on_stopped = on_stopped

    def __call__(self, f):
        @wraps(f)
        def delay(*args, **kwargs):
            from scheduler.helpers.queues import get_queue

            queue = get_queue(self.queue) if isinstance(self.queue, str) else self.queue

            job_name = kwargs.pop("job_name", None)
            at_front = kwargs.pop("at_front", False)

            if not at_front:
                at_front = self.at_front

            return queue.create_and_enqueue_job(
                f,
                args=args,
                kwargs=kwargs,
                timeout=self.timeout,
                result_ttl=self.result_ttl,
                job_info_ttl=self.job_info_ttl,
                name=job_name,
                at_front=at_front,
                meta=self.meta,
                description=self.description,
                on_failure=self.on_failure,
                on_success=self.on_success,
                on_stopped=self.on_stopped,
                when=None,
            )

        JOB_METHODS_LIST.append(f"{f.__module__}.{f.__name__}")
        f.delay = delay
        return f



import dataclasses
from collections.abc import Sequence
from typing import ClassVar, Optional, List, Tuple, Any

from scheduler.helpers.utils import current_timestamp
from scheduler.redis_models.base import as_str, BaseModel
from scheduler.settings import logger
from scheduler.types import ConnectionType, Self


class DequeueTimeout(Exception):
    pass


@dataclasses.dataclass(slots=True, kw_only=True)
class ZSetModel(BaseModel):
    def cleanup(self, connection: ConnectionType, timestamp: Optional[float] = None) -> None:
        """Remove expired jobs from registry."""
        score = timestamp or current_timestamp()
        connection.zremrangebyscore(self._key, 0, score)

    def count(self, connection: ConnectionType) -> int:
        """Returns the number of jobs in this registry"""
        self.cleanup(connection=connection)
        return connection.zcard(self._key)

    def add(self, connection: ConnectionType, job_name: str, score: float, update_existing_only: bool = False) -> int:
        return connection.zadd(self._key, {job_name: float(score)}, xx=update_existing_only)

    def delete(self, connection: ConnectionType, job_name: str) -> None:
        connection.zrem(self._key, job_name)


class JobNamesRegistry(ZSetModel):
    _element_key_template: ClassVar[str] = ":registry:{}"

    def __init__(self, connection: ConnectionType, name: str) -> None:
        super().__init__(name=name)
        self.connection = connection

    def __len__(self) -> int:
        return self.count(self.connection)

    def __contains__(self, item: str) -> bool:
        return self.connection.zrank(self._key, item) is not None

    def all(self, start: int = 0, end: int = -1) -> List[str]:
        """Returns list of all job names.

        :param start: Start score/timestamp, default to 0.
        :param end: End score/timestamp, default to -1 (i.e., no max score).
        :returns: Returns list of all job names with timestamp from start to end
        """
        self.cleanup(self.connection)
        res = [as_str(job_name) for job_name in self.connection.zrange(self._key, start, end)]
        logger.debug(f"Getting jobs for registry {self._key}: {len(res)} found.")
        return res

    def all_with_timestamps(self, start: int = 0, end: int = -1) -> List[Tuple[str, float]]:
        """Returns list of all job names with their timestamps.

        :param start: Start score/timestamp, default to 0.
        :param end: End score/timestamp, default to -1 (i.e., no max score).
        :returns: Returns list of all job names with timestamp from start to end
        """
        self.cleanup(self.connection)
        res = self.connection.zrange(self._key, start, end, withscores=True)
        logger.debug(f"Getting jobs for registry {self._key}: {len(res)} found.")
        return [(as_str(job_name), timestamp) for job_name, timestamp in res]

    def get_first(self) -> Optional[str]:
        """Returns the first job in the registry."""
        self.cleanup(self.connection)
        first_job = self.connection.zrange(self._key, 0, 0)
        return first_job[0].decode() if first_job else None

    def get_last_timestamp(self) -> Optional[float]:
        """Returns the last timestamp in the registry."""
        self.cleanup(self.connection)
        last_timestamp = self.connection.zrange(self._key, -1, -1, withscores=True)
        return last_timestamp[0][1] if last_timestamp else None

    @property
    def key(self) -> str:
        return self._key

    @classmethod
    def pop(
        cls, connection: ConnectionType, registries: Sequence[Self], timeout: Optional[int]
    ) -> Tuple[Optional[str], Optional[str]]:
        """Helper method to abstract away from some Redis API details

        :param connection: Broker connection
        :param registries: List of registries to pop from
        :param timeout: Timeout in seconds
        :raises ValueError: If timeout of 0 was passed
        :raises DequeueTimeout: BLPOP Timeout
        :returns: Tuple of registry key and job name
        """
        if timeout == 0:
            raise ValueError("Indefinite timeout not supported. Please pick a timeout value > 0")
        registry_keys = [r.key for r in registries]
        if timeout is not None:  # blocking variant
            colored_registries = ",".join(map(str, [str(registry) for registry in registry_keys]))
            logger.debug(f"Starting BZMPOP operation for queues {colored_registries} with timeout of {timeout}")
            result = connection.bzpopmin(registry_keys, timeout)
            if not result:
                logger.debug(f"BZMPOP timeout, no jobs found on queues {colored_registries}")
                raise DequeueTimeout(timeout, registry_keys)
            registry_key, job_name, timestamp = result
            return as_str(registry_key), as_str(job_name)
        else:  # non-blocking variant
            for registry_key in registry_keys:
                results: List[Any] = connection.zpopmin(registry_key)
                if results:
                    job_name, timestamp = results[0]
                    return as_str(registry_key), as_str(job_name)
            return None, None



from datetime import timedelta

import time_machine
from django.utils import timezone

from scheduler.settings import SCHEDULER_CONFIG
from scheduler.worker import create_worker
from scheduler.models import TaskType
from scheduler.tests.testtools import SchedulerBaseCase, task_factory
from scheduler.worker import WorkerScheduler


class TestWorkerScheduler(SchedulerBaseCase):
    def test_create_worker_with_scheduler__scheduler_started(self):
        SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 1
        worker = create_worker("default", name="test", burst=True, with_scheduler=True)
        worker.bootstrap()
        self.assertIsNotNone(worker.scheduler)
        worker.stop_scheduler()
        self.assertIsNone(worker.scheduler)

    def test_scheduler_schedules_tasks(self):
        with time_machine.travel(0.0, tick=False) as traveller:
            # arrange
            task = task_factory(TaskType.ONCE, scheduled_time=timezone.now() + timedelta(milliseconds=40))
            self.assertIsNotNone(task.job_name)
            self.assertNotIn(task.job_name, task.rqueue.queued_job_registry)
            self.assertIn(task.job_name, task.rqueue.scheduled_job_registry)

            scheduler = WorkerScheduler(
                [
                    task.rqueue,
                ],
                worker_name="fake-worker",
                connection=task.rqueue.connection,
            )

            # act
            traveller.move_to(50)
            scheduler._acquire_locks()
            scheduler.enqueue_scheduled_jobs()

            # assert
            self.assertIsNotNone(task.job_name)
            self.assertIn(task.job_name, task.rqueue.queued_job_registry)
            self.assertNotIn(task.job_name, task.rqueue.scheduled_job_registry)



from unittest import mock

from django.core.management import call_command
from django.test import TestCase

from scheduler.helpers.queues import get_queue
from scheduler.redis_models import JobModel
from scheduler.settings import SCHEDULER_CONFIG
from scheduler.tests import test_settings  # noqa
from scheduler.tests.jobs import failing_job


class SchedulerWorkerTestCase(TestCase):
    def test_scheduler_worker__no_queues_params(self):
        SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 1
        queue = get_queue("default")

        # enqueue some jobs that will fail
        job_names = []
        for _ in range(0, 3):
            job = queue.create_and_enqueue_job(failing_job)
            job_names.append(job.name)

        # Create a worker to execute these jobs
        call_command("scheduler_worker", fork_job_execution=False, burst=True)

        # check if all jobs are really failed
        for job_name in job_names:
            job = JobModel.get(name=job_name, connection=queue.connection)
            self.assertTrue(job.is_failed)
        SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 10

    @mock.patch("scheduler.management.commands.scheduler_worker.create_worker")
    def test_scheduler_worker__without_scheduler(self, mock_create_worker):
        # Create a worker to execute these jobs
        call_command("scheduler_worker", "default", "--burst", "--without-scheduler")
        mock_create_worker.assert_called_once_with(
            "default",
            name=None,
            fork_job_execution=True,
            burst=True,
            with_scheduler=False,
        )

    @mock.patch("scheduler.management.commands.scheduler_worker.create_worker")
    def test_scheduler_worker__with_scheduler(self, mock_create_worker):
        # Create a worker to execute these jobs
        call_command("scheduler_worker", "default", "--burst")
        mock_create_worker.assert_called_once_with(
            "default",
            name=None,
            fork_job_execution=True,
            burst=True,
            with_scheduler=True,
        )

    def test_scheduler_worker__run_jobs(self):
        SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 1
        queue = get_queue("default")

        # enqueue some jobs that will fail
        job_names = []
        for _ in range(0, 3):
            job = queue.create_and_enqueue_job(failing_job)
            job_names.append(job.name)

        # Create a worker to execute these jobs
        call_command("scheduler_worker", "default", fork_job_execution=False, burst=True)

        # check if all jobs are really failed
        for job_name in job_names:
            job = JobModel.get(name=job_name, connection=queue.connection)
            self.assertTrue(job.is_failed)
        SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 10

    def test_scheduler_worker__worker_with_two_queues(self):
        SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 1
        queue = get_queue("default")
        queue2 = get_queue("django_tasks_scheduler_test")

        # enqueue some jobs that will fail
        job_names = []
        for _ in range(0, 3):
            job = queue.create_and_enqueue_job(failing_job)
            job_names.append(job.name)
        job = queue2.create_and_enqueue_job(failing_job)
        job_names.append(job.name)

        # Create a worker to execute these jobs
        call_command("scheduler_worker", "default", "django_tasks_scheduler_test", fork_job_execution=False, burst=True)

        # check if all jobs are really failed
        for job_name in job_names:
            job = JobModel.get(name=job_name, connection=queue.connection)
            self.assertTrue(job.is_failed)
        SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 10

    def test_scheduler_worker__worker_with_one_queue__does_not_perform_other_queue_job(self):
        SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 1
        queue = get_queue("default")
        queue2 = get_queue("django_tasks_scheduler_test")

        job = queue.create_and_enqueue_job(failing_job)
        other_job = queue2.create_and_enqueue_job(failing_job)

        # Create a worker to execute these jobs
        call_command("scheduler_worker", "default", fork_job_execution=False, burst=True)

        # assert
        job = JobModel.get(job.name, connection=queue.connection)
        self.assertTrue(job.is_failed)
        other_job = JobModel.get(other_job.name, connection=queue.connection)

        self.assertTrue(other_job.is_queued, f"Expected other job to be queued but status={other_job.status}")
        SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 10



import json
from abc import ABC
from datetime import datetime, timezone
from typing import Type, Dict, Any

from scheduler.settings import logger
from scheduler.types import ConnectionType, Self

_PUBSUB_CHANNEL_TEMPLATE: str = ":workers:pubsub:{}"


class WorkerCommandError(Exception):
    pass


class WorkerCommand(ABC):
    """Abstract class for commands to be sent to a worker and processed by worker"""

    _registry: Dict[str, Type[Self]] = dict()
    command_name: str = ""

    def __init__(self, *args, worker_name: str, **kwargs) -> None:
        self.worker_name = worker_name

    def command_payload(self, **kwargs) -> Dict[str, Any]:
        commands_channel = WorkerCommandsChannelListener._commands_channel(self.worker_name)
        payload = {
            "command": self.command_name,
            "worker_name": self.worker_name,
            "channel_name": commands_channel,
            "created_at": datetime.now(tz=timezone.utc).isoformat(),
        }
        if kwargs:
            payload.update(kwargs)
        return payload

    def __str__(self) -> str:
        return f"{self.command_name}[{self.command_payload()}]"

    def process_command(self, connection: ConnectionType) -> None:
        raise NotImplementedError

    @classmethod
    def __init_subclass__(cls, *args, **kwargs):
        if cls is WorkerCommand:
            return
        if not cls.command_name:
            raise NotImplementedError(f"{cls.__name__} must have a name attribute")
        WorkerCommand._registry[cls.command_name] = cls

    @classmethod
    def from_payload(cls, payload: Dict[str, Any]) -> Type[Self]:
        command_name = payload.get("command")
        command_class = WorkerCommand._registry.get(command_name)
        if command_class is None:
            raise WorkerCommandError(f"Invalid command: {command_name}")
        return command_class(**payload)


def send_command(connection: ConnectionType, command: WorkerCommand) -> None:
    """Send a command to the worker"""
    payload = command.command_payload()
    connection.publish(payload["channel_name"], json.dumps(payload))


class WorkerCommandsChannelListener(object):
    def __init__(self, connection: ConnectionType, worker_name: str) -> None:
        self.connection = connection
        self.pubsub_channel_name = WorkerCommandsChannelListener._commands_channel(worker_name)

    @staticmethod
    def _commands_channel(worker_name: str) -> str:
        return _PUBSUB_CHANNEL_TEMPLATE.format(worker_name)

    def start(self):
        """Subscribe to this worker's channel"""
        logger.info(f"Subscribing to channel {self.pubsub_channel_name}")
        self.pubsub = self.connection.pubsub()
        self.pubsub.subscribe(**{self.pubsub_channel_name: self.handle_payload})
        self.pubsub_thread = self.pubsub.run_in_thread(sleep_time=0.2, daemon=True)

    def stop(self):
        """Unsubscribe from pubsub channel"""
        if self.pubsub_thread:
            logger.info(f"Unsubscribing from channel {self.pubsub_channel_name}")
            self.pubsub_thread.stop()
            self.pubsub_thread.join()
            self.pubsub.unsubscribe()
            self.pubsub.close()

    def handle_payload(self, payload: str) -> None:
        """Handle commands"""
        command = WorkerCommand.from_payload(json.loads(payload["data"]))
        logger.debug(f"Received command: {command}")
        command.process_command(self.connection)



import logging
from typing import List, Dict

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from scheduler.types import SchedulerConfiguration, QueueConfiguration

logger = logging.getLogger("scheduler")
logging.basicConfig(level=logging.DEBUG)

_QUEUES: Dict[str, QueueConfiguration] = dict()
SCHEDULER_CONFIG: SchedulerConfiguration = SchedulerConfiguration()


class QueueNotFoundError(Exception):
    pass


def conf_settings():
    global _QUEUES
    global SCHEDULER_CONFIG

    app_queues = getattr(settings, "SCHEDULER_QUEUES", None)
    if app_queues is None or not isinstance(app_queues, dict):
        raise ImproperlyConfigured("You have to define SCHEDULER_QUEUES in settings.py as dict")

    for queue_name, queue_config in app_queues.items():
        if isinstance(queue_config, QueueConfiguration):
            _QUEUES[queue_name] = queue_config
        elif isinstance(queue_config, dict):
            _QUEUES[queue_name] = QueueConfiguration(**queue_config)
        else:
            raise ImproperlyConfigured(f"Queue {queue_name} configuration should be a QueueConfiguration or dict")

    user_settings = getattr(settings, "SCHEDULER_CONFIG", {})
    if isinstance(user_settings, SchedulerConfiguration):
        return
    if not isinstance(user_settings, dict):
        raise ImproperlyConfigured("SCHEDULER_CONFIG should be a SchedulerConfiguration or dict")
    for k in user_settings:
        if k not in SCHEDULER_CONFIG.__annotations__:
            raise ImproperlyConfigured(f"Unknown setting {k} in SCHEDULER_CONFIG")
        setattr(SCHEDULER_CONFIG, k, getattr(user_settings, k, None))


conf_settings()


def get_queue_names() -> List[str]:
    return list(_QUEUES.keys())


def get_queue_configuration(queue_name: str) -> QueueConfiguration:
    if queue_name not in _QUEUES:
        raise QueueNotFoundError(f"Queue {queue_name} not found, queues={_QUEUES.keys()}")
    return _QUEUES[queue_name]



__all__ = [
    "job_detail",
    "job_action",
    "stats",
    "stats_json",
    "queue_registry_actions",
    "queue_confirm_job_action",
    "queue_workers",
    "queue_job_actions",
    "list_registry_jobs",
    "workers_list",
    "worker_details",
    "get_statistics",
]

from .job_views import job_detail, job_action
from .queue_job_actions import queue_job_actions, queue_confirm_job_action
from .queue_registry_actions import queue_registry_actions
from .queue_views import (
    stats,
    stats_json,
    queue_workers,
    list_registry_jobs,
    get_statistics,
)
from .worker_views import workers_list, worker_details



__all__ = [
    "ConnectionErrorTypes",
    "ResponseErrorTypes",
    "TimeoutErrorTypes",
    "WatchErrorTypes",
    "ConnectionType",
    "PipelineType",
    "SentinelType",
    "FunctionReferenceType",
    "BrokerMetaData",
    "TASK_TYPES",
    "Broker",
    "SchedulerConfiguration",
    "QueueConfiguration",
    "Self",
]

from .broker_types import (
    ConnectionErrorTypes,
    ResponseErrorTypes,
    TimeoutErrorTypes,
    WatchErrorTypes,
    ConnectionType,
    PipelineType,
    SentinelType,
    FunctionReferenceType,
    BrokerMetaData,
    TASK_TYPES,
)
from .settings_types import Broker, SchedulerConfiguration, QueueConfiguration, Self



"""list_registry_jobs actions on all jobs in the registry"""

from enum import Enum

from django.contrib import admin, messages
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse, HttpRequest, HttpResponseNotFound
from django.shortcuts import render, redirect
from django.urls import reverse
from django.views.decorators.cache import never_cache

from scheduler.helpers.queues import Queue
from scheduler.redis_models import JobModel, JobNamesRegistry
from scheduler.types import ResponseErrorTypes
from scheduler.views.helpers import get_queue, _check_next_url, _enqueue_multiple_jobs


class QueueRegistryActions(Enum):
    EMPTY = "empty"
    REQUEUE = "requeue"


def _clear_registry(request: HttpRequest, queue: Queue, registry_name: str, registry: JobNamesRegistry):
    try:
        job_names = registry.all()
        for job_name in job_names:
            registry.delete(registry.connection, job_name)
            job_model = JobModel.get(job_name, connection=registry.connection)
            job_model.delete(connection=registry.connection)
        messages.info(request, f"You have successfully cleared the {registry_name} jobs in queue {queue.name}")
    except ResponseErrorTypes as e:
        messages.error(request, f"error: {e}")
        raise e


def _requeue_job_names(request: HttpRequest, queue: Queue, registry_name: str):
    registry = queue.get_registry(registry_name)
    job_names = registry.all()
    jobs_requeued_count = _enqueue_multiple_jobs(queue, job_names)
    messages.info(request, f"You have successfully re-queued {jobs_requeued_count} jobs!")


@never_cache
@staff_member_required
def queue_registry_actions(request: HttpRequest, queue_name: str, registry_name: str, action: str) -> HttpResponse:
    queue = get_queue(queue_name)
    registry = queue.get_registry(registry_name)
    if registry is None:
        return HttpResponseNotFound()
    next_url = _check_next_url(request, reverse("queue_registry_jobs", args=[queue_name, registry_name]))
    if action not in [item.value for item in QueueRegistryActions]:
        return redirect(next_url)
    if request.method == "POST":
        if action == QueueRegistryActions.EMPTY.value:
            _clear_registry(request, queue, registry_name, registry)
        elif action == QueueRegistryActions.REQUEUE.value:
            _requeue_job_names(request, queue, registry_name)
        return redirect("queue_registry_jobs", queue_name, registry_name)
    job_names = registry.all()
    job_list = JobModel.get_many(job_names, connection=queue.connection)
    context_data = {
        **admin.site.each_context(request),
        "queue": queue,
        "total_jobs": len(registry),
        "action": action,
        "jobs": job_list,
        "next_url": next_url,
        "action_url": reverse("queue_registry_action", args=[queue_name, registry_name, action]),
    }
    return render(request, "admin/scheduler/confirm_action.html", context_data)



import os
import uuid

from scheduler import settings
from scheduler.redis_models import WorkerModel
from scheduler.worker import create_worker
from scheduler.tests import test_settings  # noqa
from scheduler.tests.testtools import SchedulerBaseCase
from scheduler.worker.worker import QueueConnectionDiscrepancyError


class TestWorker(SchedulerBaseCase):
    def test_create_worker__two_workers_same_queue(self):
        worker1 = create_worker("default", "django_tasks_scheduler_test")
        worker1.worker_start()
        worker2 = create_worker("default")
        worker2.worker_start()
        hostname = os.uname()[1]
        self.assertEqual(f"{hostname}-worker.1", worker1.name)
        self.assertEqual(f"{hostname}-worker.2", worker2.name)

    def test_create_worker__worker_with_queues_different_connection(self):
        with self.assertRaises(QueueConnectionDiscrepancyError):
            create_worker("default", "test1")

    def test_create_worker__with_name(self):
        name = uuid.uuid4().hex
        worker1 = create_worker("default", name=name)
        self.assertEqual(name, worker1.name)

    def test_create_worker__with_name_containing_slash(self):
        name = uuid.uuid4().hex[-4:] + "/" + uuid.uuid4().hex[-4:]
        worker1 = create_worker("default", name=name)
        self.assertEqual(name.replace("/", "."), worker1.name)

    def test_create_worker__scheduler_interval(self):
        prev = settings.SCHEDULER_CONFIG.SCHEDULER_INTERVAL
        settings.SCHEDULER_CONFIG.SCHEDULER_INTERVAL = 1
        worker = create_worker("default", name="test", burst=True, with_scheduler=True)
        worker.bootstrap()
        self.assertEqual(worker.name, "test")
        self.assertEqual(worker.scheduler.interval, 1)
        settings.SCHEDULER_CONFIG.SCHEDULER_INTERVAL = prev
        worker.teardown()

    def test_create_worker__cleanup(self):
        worker = create_worker("default", name="test", burst=True, with_scheduler=False)
        worker.bootstrap()
        worker.connection.delete(WorkerModel.key_for(worker.name))
        all_names = WorkerModel.all_names(worker.connection)
        self.assertIn(worker.name, all_names)
        # act
        WorkerModel.cleanup(worker.connection, "default")
        # assert
        all_names = WorkerModel.all_names(worker.connection)
        self.assertNotIn(worker.name, all_names)



import os
import time
import traceback
from datetime import datetime
from enum import Enum
from threading import Thread
from typing import List, Set, Optional, Sequence, Dict

import django

from scheduler.helpers.queues import Queue
from scheduler.helpers.queues import get_queue
from scheduler.helpers.utils import current_timestamp
from scheduler.models import Task
from scheduler.redis_models import SchedulerLock, JobModel, ScheduledJobRegistry
from scheduler.settings import SCHEDULER_CONFIG, logger
from scheduler.types import ConnectionType


class SchedulerStatus(str, Enum):
    STARTED = "started"
    WORKING = "working"
    STOPPED = "stopped"


def _reschedule_tasks():
    enabled_jobs = list(Task.objects.filter(enabled=True))
    for item in enabled_jobs:
        logger.debug(f"Rescheduling {str(item)}")
        item.save()


class WorkerScheduler:
    def __init__(
        self,
        queues: Sequence[Queue],
        connection: ConnectionType,
        worker_name: str,
        interval: Optional[int] = None,
    ) -> None:
        interval = interval or SCHEDULER_CONFIG.SCHEDULER_INTERVAL
        self._queues = queues
        self._scheduled_job_registries: List[ScheduledJobRegistry] = []
        self.lock_acquisition_time = None
        self._pool_class = connection.connection_pool.connection_class
        self._pool_kwargs = connection.connection_pool.connection_kwargs.copy()
        self._locks: Dict[str, SchedulerLock] = dict()
        self.connection = connection
        self.interval = interval
        self._stop_requested = False
        self.status = SchedulerStatus.STOPPED
        self._thread = None
        self._pid: Optional[int] = None
        self.worker_name = worker_name

    @property
    def pid(self) -> Optional[int]:
        return self._pid

    def _should_reacquire_locks(self) -> bool:
        """Returns True if lock_acquisition_time is longer than 10 minutes ago"""
        if not self.lock_acquisition_time:
            return True
        seconds_since = (datetime.now() - self.lock_acquisition_time).total_seconds()
        return seconds_since > SCHEDULER_CONFIG.SCHEDULER_FALLBACK_PERIOD_SECS

    def _acquire_locks(self) -> Set[str]:
        """Returns names of queue it successfully acquires lock on"""
        successful_locks = set()
        if self.pid is None:
            self._pid = os.getpid()
        queue_names = [queue.name for queue in self._queues]
        logger.debug(
            f"""[Scheduler {self.worker_name}/{self.pid}] Trying to acquire locks for {", ".join(queue_names)}"""
        )
        for queue in self._queues:
            lock = SchedulerLock(queue.name)
            if lock.acquire(self.pid, connection=queue.connection, expire=self.interval + 60):
                self._locks[queue.name] = lock
                successful_locks.add(queue.name)

        # Always reset _scheduled_job_registries when acquiring locks
        self.lock_acquisition_time = datetime.now()
        self._scheduled_job_registries = []
        for queue_name in self._locks:
            queue = get_queue(queue_name)
            self._scheduled_job_registries.append(queue.scheduled_job_registry)
        logger.debug(f"[Scheduler {self.worker_name}/{self.pid}] Locks acquired for {', '.join(self._locks.keys())}")
        return successful_locks

    def start(self) -> None:
        locks = self._acquire_locks()
        if len(locks) == 0:
            return
        self.status = SchedulerStatus.STARTED
        self._thread = Thread(target=run_scheduler, args=(self,), name="scheduler-thread")
        self._thread.start()

    def request_stop_and_wait(self):
        """Toggle self._stop_requested that's checked on every loop"""
        logger.debug(f"[Scheduler {self.worker_name}/{self.pid}] Stop Scheduler requested")
        self._stop_requested = True
        if self._thread is not None:
            self._thread.join()

    def heartbeat(self):
        """Updates the TTL on scheduler keys and the locks"""
        lock_keys = ", ".join(self._locks.keys())
        logger.debug(f"[Scheduler {self.worker_name}/{self.pid}] Scheduler updating lock for queue {lock_keys}")
        with self.connection.pipeline() as pipeline:
            for lock in self._locks.values():
                lock.expire(self.connection, expire=self.interval + 60)
            pipeline.execute()

    def stop(self):
        logger.info(
            f"[Scheduler {self.worker_name}/{self.pid}] Stopping scheduler, releasing locks for {', '.join(self._locks.keys())}..."
        )
        self.release_locks()
        self.status = SchedulerStatus.STOPPED

    def release_locks(self):
        """Release acquired locks"""
        with self.connection.pipeline() as pipeline:
            for lock in self._locks.values():
                lock.release(self.connection)
            pipeline.execute()

    def work(self) -> None:
        queue_names = [queue.name for queue in self._queues]
        logger.info(f"""[Scheduler {self.worker_name}/{self.pid}] Scheduler for {", ".join(queue_names)} started""")
        django.setup()

        while True:
            if self._stop_requested:
                self.stop()
                break

            if self._should_reacquire_locks():
                self._acquire_locks()

            self.enqueue_scheduled_jobs()
            self.heartbeat()
            time.sleep(self.interval)

    def enqueue_scheduled_jobs(self) -> None:
        """Enqueue jobs whose timestamp is in the past"""
        self.status = SchedulerStatus.WORKING
        _reschedule_tasks()

        for registry in self._scheduled_job_registries:
            timestamp = current_timestamp()
            job_names = registry.get_jobs_to_schedule(timestamp)
            if len(job_names) == 0:
                continue
            queue = get_queue(registry.name)
            jobs = JobModel.get_many(job_names, connection=self.connection)
            with self.connection.pipeline() as pipeline:
                for job in jobs:
                    if job is not None:
                        queue.enqueue_job(job, connection=pipeline, at_front=job.at_front)
                pipeline.execute()
        self.status = SchedulerStatus.STARTED


def run_scheduler(scheduler: WorkerScheduler):
    try:
        scheduler.work()
    except:  # noqa
        logger.error(f"Scheduler [PID {os.getpid()}] raised an exception.\n{traceback.format_exc()}")
        raise
    logger.info(f"Scheduler with PID {os.getpid()} has stopped")



import sys

import click
from django.core.management.base import BaseCommand

from scheduler.models import Task


class Command(BaseCommand):
    """Export all scheduled jobs"""

    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "-o",
            "--output",
            action="store",
            choices=["json", "yaml"],
            default="json",
            dest="format",
            help="format of output",
        )

        parser.add_argument(
            "-e",
            "--enabled",
            action="store_true",
            dest="enabled",
            help="Export only enabled jobs",
        )
        parser.add_argument(
            "-f",
            "--filename",
            action="store",
            dest="filename",
            help="File name to load (otherwise writes to standard output)",
        )

    def handle(self, *args, **options):
        file = open(options.get("filename"), "w") if options.get("filename") else sys.stdout
        res = list()

        tasks = Task.objects.all()
        if options.get("enabled"):
            tasks = tasks.filter(enabled=True)
        for task in tasks:
            res.append(task.to_dict())

        if options.get("format") == "json":
            import json

            click.echo(json.dumps(res, indent=2, default=str), file=file)
            return

        if options.get("format") == "yaml":
            try:
                import yaml
            except ImportError:
                click.echo("Aborting. LibYAML is not installed.")
                exit(1)
            # Disable YAML alias
            yaml.Dumper.ignore_aliases = lambda *x: True
            click.echo(yaml.dump(res, default_flow_style=False), file=file)
            return



__all__ = [
    "Result",
    "ResultType",
    "as_str",
    "SchedulerLock",
    "WorkerModel",
    "DequeueTimeout",
    "KvLock",
    "JobStatus",
    "JobModel",
    "JobNamesRegistry",
    "FinishedJobRegistry",
    "ActiveJobRegistry",
    "FailedJobRegistry",
    "CanceledJobRegistry",
    "ScheduledJobRegistry",
    "QueuedJobRegistry",
]

from .base import as_str
from .job import JobStatus, JobModel
from .lock import SchedulerLock, KvLock
from .registry.base_registry import DequeueTimeout, JobNamesRegistry
from .registry.queue_registries import (
    FinishedJobRegistry,
    ActiveJobRegistry,
    FailedJobRegistry,
    CanceledJobRegistry,
    ScheduledJobRegistry,
    QueuedJobRegistry,
)
from .result import Result, ResultType
from .worker import WorkerModel



import uuid
from datetime import datetime
from unittest.mock import patch, PropertyMock

from django.urls import reverse

from scheduler.helpers.queues import get_queue
from scheduler.models import TaskType
from scheduler.redis_models import JobModel, WorkerModel
from scheduler.tests import test_settings  # noqa
from scheduler.tests.jobs import failing_job, test_job
from scheduler.tests.test_task_types.test_task_model import assert_response_has_msg
from scheduler.tests.test_views.base import BaseTestCase
from scheduler.tests.testtools import assert_message_in_response, task_factory, _get_task_scheduled_job_from_registry
from scheduler.types import QueueConfiguration
from scheduler.worker import create_worker


class TestViewJobDetails(BaseTestCase):
    def test_job_details(self):
        """Job data is displayed properly"""
        queue = get_queue("default")
        job = queue.create_and_enqueue_job(test_job)

        url = reverse("job_details", args=[job.name])
        res = self.client.get(url)
        self.assertEqual(200, res.status_code)
        self.assertIn("job", res.context)
        self.assertEqual(res.context["job"], job)

        # Bad job-id should return 404
        url = reverse("job_details", args=["bad_job_name"])
        res = self.client.get(url, follow=True)
        self.assertEqual(200, res.status_code)
        assert_response_has_msg(res, "Job bad_job_name does not exist, maybe its TTL has passed")

    def test_scheduled_job_details(self):
        """Job data is displayed properly"""
        scheduled_job = task_factory(TaskType.ONCE, enabled=True)
        job = _get_task_scheduled_job_from_registry(scheduled_job)

        url = reverse(
            "job_details",
            args=[
                job.name,
            ],
        )
        res = self.client.get(url, follow=True)
        self.assertIn("job", res.context)
        self.assertEqual(res.context["job"], job)

    def test_job_details_on_deleted_dependency(self):
        """Page doesn't crash even if job.dependency has been deleted"""
        queue = get_queue("default")

        job = queue.create_and_enqueue_job(test_job)
        second_job = queue.create_and_enqueue_job(test_job)
        queue.delete_job(job.name)
        url = reverse("job_details", args=[second_job.name])
        res = self.client.get(url)
        self.assertEqual(res.status_code, 200)

    def test_requeue_all(self):
        """Ensure that re-queuing all failed job work properly"""
        queue = get_queue("default")
        queue_name = "default"
        queue.create_and_enqueue_job(failing_job)
        queue.create_and_enqueue_job(failing_job)
        worker = create_worker("default", burst=True)
        worker.work()

        res = self.client.get(reverse("queue_registry_action", args=[queue_name, "failed", "requeue"]))
        self.assertEqual(res.context["total_jobs"], 2)
        # After requeue_all is called, jobs are enqueued
        res = self.client.post(reverse("queue_registry_action", args=[queue_name, "failed", "requeue"]))
        self.assertEqual(len(queue), 4)

    def test_requeue_all_if_deleted_job(self):
        """
        Ensure that re-queuing all failed job work properly
        """
        queue = get_queue("default")
        queue_name = "default"
        job = queue.create_and_enqueue_job(failing_job)
        queue.create_and_enqueue_job(failing_job)
        worker = create_worker("default", burst=True)
        worker.work()

        res = self.client.get(reverse("queue_registry_action", args=[queue_name, "failed", "requeue"]))
        self.assertEqual(res.context["total_jobs"], 2)
        queue.delete_job(job.name)

        # After requeue_all is called, jobs are enqueued
        res = self.client.post(reverse("queue_registry_action", args=[queue_name, "failed", "requeue"]))
        self.assertEqual(len(queue.queued_job_registry), 1)

    def test_clear_queue_unknown_registry(self):
        queue_name = "django_tasks_scheduler_test"
        res = self.client.post(reverse("queue_registry_action", args=[queue_name, "unknown", "empty"]), {"post": "yes"})
        self.assertEqual(404, res.status_code)

    def test_clear_queue_enqueued(self):
        queue = get_queue("django_tasks_scheduler_test")
        job = queue.create_and_enqueue_job(test_job)
        self.client.post(reverse("queue_registry_action", args=[queue.name, "queued", "empty"]), {"post": "yes"})
        self.assertFalse(JobModel.exists(job.name, connection=queue.connection), f"job {job.name} exists")
        self.assertNotIn(job.name, queue.queued_job_registry.all())

    def test_clear_queue_scheduled(self):
        queue = get_queue("django_tasks_scheduler_test")
        job = queue.create_and_enqueue_job(test_job, when=datetime.now())

        res = self.client.get(reverse("queue_registry_action", args=[queue.name, "scheduled", "empty"]), follow=True)
        self.assertEqual(200, res.status_code)
        self.assertEqual(res.context["jobs"], [job])

        res = self.client.post(
            reverse("queue_registry_action", args=[queue.name, "scheduled", "empty"]), {"post": "yes"}, follow=True
        )
        assert_message_in_response(res, f"You have successfully cleared the scheduled jobs in queue {queue.name}")
        self.assertEqual(200, res.status_code)
        self.assertFalse(JobModel.exists(job.name, connection=queue.connection))
        self.assertNotIn(job.name, queue.queued_job_registry.all())

    def test_queue_workers(self):
        """Worker index page should show workers for a specific queue"""
        queue_name = "django_tasks_scheduler_test"

        worker1 = create_worker(queue_name)
        worker1.worker_start()
        worker2 = create_worker("test3")
        worker2.worker_start()

        res = self.client.get(reverse("queue_workers", args=[queue_name]))
        worker1_model = WorkerModel.get(worker1.name, connection=worker1.connection)
        self.assertEqual(res.context["workers"], [worker1_model])

    def test_worker_details(self):
        """Worker index page should show workers for a specific queue"""

        worker = create_worker("django_tasks_scheduler_test", name=uuid.uuid4().hex)
        worker.worker_start()

        url = reverse("worker_details", args=[worker.name])
        res = self.client.get(url)
        self.assertEqual(res.context["worker"], worker._model)

    def test_worker_details__non_existing_worker(self):
        """Worker index page should show workers for a specific queue"""

        worker = create_worker("django_tasks_scheduler_test", name="WORKER")
        worker.worker_start()

        res = self.client.get(reverse("worker_details", args=["bad-worker-name"]))
        self.assertEqual(404, res.status_code)

    def test_statistics_json_view(self):
        # Override testing SCHEDULER_QUEUES
        queues = {
            "default": QueueConfiguration(DB=0, HOST="localhost", PORT=6379),
        }
        with patch("scheduler.settings._QUEUES", new_callable=PropertyMock(return_value=queues)):
            res = self.client.get(reverse("queues_home"))
            self.assertEqual(res.status_code, 200)

            res = self.client.get(reverse("queues_home_json"))
            self.assertEqual(res.status_code, 200)

            # Not staff => return 404
            self.user.is_staff = False
            self.user.save()

            res = self.client.get(reverse("queues_home"))
            self.assertEqual(res.status_code, 302)

            # 404 code for stats
            res = self.client.get(reverse("queues_home_json"))
            self.assertEqual(res.status_code, 404)

    @staticmethod
    def token_validation(token: str) -> bool:
        return token == "valid"

    # @patch('scheduler.views.SCHEDULER_CONFIG')
    # def test_statistics_json_view_token(self, configuration):
    #     configuration.get.return_value = ViewTest.token_validation
    #     self.user.is_staff = False
    #     self.user.save()
    #     res = self.client.get(reverse('queues_home_json'), headers={'Authorization': 'valid'})
    #     self.assertEqual(res.status_code, 200)
    #
    #     res = self.client.get(reverse('queues_home_json'), headers={'Authorization': 'invalid'})
    #     self.assertEqual(res.status_code, 404)



import asyncio
import sys
import traceback
from datetime import datetime
from typing import Dict, List, Optional, Tuple, Union, Any

from redis import WatchError

from scheduler.helpers.callback import Callback
from scheduler.helpers.utils import utcnow, current_timestamp
from scheduler.redis_models import (
    JobNamesRegistry,
    FinishedJobRegistry,
    ActiveJobRegistry,
    FailedJobRegistry,
    CanceledJobRegistry,
    ScheduledJobRegistry,
    QueuedJobRegistry,
)
from scheduler.redis_models import JobStatus, SchedulerLock, Result, ResultType, JobModel
from scheduler.settings import logger, SCHEDULER_CONFIG
from scheduler.types import ConnectionType, FunctionReferenceType, Self


class InvalidJobOperation(Exception):
    pass


class NoSuchJobError(Exception):
    pass


def perform_job(job_model: JobModel, connection: ConnectionType) -> Any:  # noqa
    """The main execution method. Invokes the job function with the job arguments.

    :returns: The job's return value
    """
    job_model.persist(connection=connection)
    _job_stack.append(job_model)

    try:
        result = job_model.func(*job_model.args, **job_model.kwargs)
        if asyncio.iscoroutine(result):
            loop = asyncio.new_event_loop()
            coro_result = loop.run_until_complete(result)
            result = coro_result
        if job_model.success_callback:
            job_model.success_callback(job_model, connection, result)  # type: ignore
        return result
    except:
        if job_model.failure_callback:
            job_model.failure_callback(job_model, connection, *sys.exc_info())  # type: ignore
        raise
    finally:
        assert job_model is _job_stack.pop()


_job_stack = []


class Queue:
    REGISTRIES = dict(
        finished="finished_job_registry",
        failed="failed_job_registry",
        scheduled="scheduled_job_registry",
        active="active_job_registry",
        canceled="canceled_job_registry",
        queued="queued_job_registry",
    )

    def __init__(self, connection: Optional[ConnectionType], name: str, is_async: bool = True) -> None:
        """Initializes a Queue object.

        :param name: The queue name
        :param connection: Broker connection
        :param is_async: Whether jobs should run "async" (using the worker).
        """
        self.connection = connection
        self.name = name
        self._is_async = is_async
        self.queued_job_registry = QueuedJobRegistry(connection=self.connection, name=self.name)
        self.active_job_registry = ActiveJobRegistry(connection=self.connection, name=self.name)
        self.failed_job_registry = FailedJobRegistry(connection=self.connection, name=self.name)
        self.finished_job_registry = FinishedJobRegistry(connection=self.connection, name=self.name)
        self.scheduled_job_registry = ScheduledJobRegistry(connection=self.connection, name=self.name)
        self.canceled_job_registry = CanceledJobRegistry(connection=self.connection, name=self.name)

    def __len__(self):
        return self.count

    @property
    def scheduler_pid(self) -> int:
        lock = SchedulerLock(self.name)
        pid = lock.value(self.connection)
        return int(pid.decode()) if pid is not None else None

    def clean_registries(self, timestamp: Optional[float] = None) -> None:
        """Remove abandoned jobs from registry and add them to FailedJobRegistry.

        Removes jobs with an expiry time earlier than current_timestamp, specified as seconds since the Unix epoch.
        Removed jobs are added to the global failed job queue.
        """
        before_score = timestamp or current_timestamp()
        self.queued_job_registry.compact()
        started_jobs: List[Tuple[str, float]] = self.active_job_registry.get_job_names_before(
            self.connection, before_score
        )

        with self.connection.pipeline() as pipeline:
            for job_name, job_score in started_jobs:
                job = JobModel.get(job_name, connection=self.connection)
                if job is None or job.failure_callback is None or job_score + job.timeout > before_score:
                    continue

                logger.debug(f"Running failure callbacks for {job.name}")
                try:
                    job.failure_callback(job, self.connection, traceback.extract_stack())
                except Exception:  # noqa
                    logger.exception(f"Job {self.name}: error while executing failure callback")
                    raise

                else:
                    logger.warning(
                        f"Queue cleanup: Moving job to {self.failed_job_registry.key} (due to AbandonedJobError)"
                    )
                    exc_string = (
                        f"Moved to {self.failed_job_registry.key}, due to AbandonedJobError, at {datetime.now()}"
                    )
                    job.status = JobStatus.FAILED
                    score = current_timestamp() + SCHEDULER_CONFIG.DEFAULT_FAILURE_TTL
                    Result.create(
                        connection=pipeline,
                        job_name=job.name,
                        worker_name=job.worker_name,
                        _type=ResultType.FAILED,
                        ttl=SCHEDULER_CONFIG.DEFAULT_FAILURE_TTL,
                        exc_string=exc_string,
                    )
                    self.failed_job_registry.add(pipeline, job.name, score)
                    job.expire(connection=pipeline, ttl=SCHEDULER_CONFIG.DEFAULT_FAILURE_TTL)
                    job.save(connection=pipeline)

            for registry in self.REGISTRIES.values():
                getattr(self, registry).cleanup(connection=self.connection, timestamp=before_score)
            pipeline.execute()

    def first_queued_job_name(self) -> Optional[str]:
        return self.queued_job_registry.get_first()

    @property
    def count(self) -> int:
        """Returns a count of all messages in the queue."""
        res = 0
        for registry in self.REGISTRIES.values():
            res += getattr(self, registry).count(connection=self.connection)
        return res

    def get_registry(self, name: str) -> Union[None, JobNamesRegistry]:
        name = name.lower()
        if name in Queue.REGISTRIES:
            return getattr(self, Queue.REGISTRIES[name])
        return None

    def get_all_job_names(self) -> List[str]:
        res = list()
        res.extend(self.queued_job_registry.all())
        res.extend(self.finished_job_registry.all())
        res.extend(self.active_job_registry.all())
        res.extend(self.failed_job_registry.all())
        res.extend(self.scheduled_job_registry.all())
        res.extend(self.canceled_job_registry.all())
        return res

    def get_all_jobs(self) -> List[JobModel]:
        job_names = self.get_all_job_names()
        return JobModel.get_many(job_names, connection=self.connection)

    def create_and_enqueue_job(
        self,
        func: FunctionReferenceType,
        args: Union[Tuple, List, None] = None,
        kwargs: Optional[Dict] = None,
        when: Optional[datetime] = None,
        timeout: Optional[int] = None,
        result_ttl: Optional[int] = None,
        job_info_ttl: Optional[int] = None,
        description: Optional[str] = None,
        name: Optional[str] = None,
        at_front: bool = False,
        meta: Optional[Dict] = None,
        on_success: Optional[Callback] = None,
        on_failure: Optional[Callback] = None,
        on_stopped: Optional[Callback] = None,
        task_type: Optional[str] = None,
        scheduled_task_id: Optional[int] = None,
        pipeline: Optional[ConnectionType] = None,
    ) -> JobModel:
        """Creates a job to represent the delayed function call and enqueues it.
        :param when: When to schedule the job (None to enqueue immediately)
        :param func: The reference to the function
        :param args: The `*args` to pass to the function
        :param kwargs: The `**kwargs` to pass to the function
        :param timeout: Function timeout
        :param result_ttl: Result time to live
        :param job_info_ttl: Time to live
        :param description: The job description
        :param name: The job name
        :param at_front: Whether to enqueue the job at the front
        :param meta: Metadata to attach to the job
        :param on_success: Callback for on success
        :param on_failure: Callback for on failure
        :param on_stopped: Callback for on stopped
        :param task_type: The task type
        :param scheduled_task_id: The scheduled task id
        :param pipeline: The Broker Pipeline
        :returns: The enqueued Job
        """
        status = JobStatus.QUEUED if when is None else JobStatus.SCHEDULED
        job_model = JobModel.create(
            connection=self.connection,
            func=func,
            args=args,
            kwargs=kwargs,
            result_ttl=result_ttl,
            job_info_ttl=job_info_ttl,
            description=description,
            name=name,
            meta=meta,
            status=status,
            timeout=timeout,
            on_success=on_success,
            on_failure=on_failure,
            on_stopped=on_stopped,
            queue_name=self.name,
            task_type=task_type,
            scheduled_task_id=scheduled_task_id,
        )
        if when is None:
            job_model = self.enqueue_job(job_model, connection=pipeline, at_front=at_front)
        elif isinstance(when, datetime):
            job_model.save(connection=self.connection)
            self.scheduled_job_registry.schedule(self.connection, job_model.name, when)
        else:
            raise TypeError(f"Invalid type for when=`{when}`")
        return job_model

    def job_handle_success(
        self, job: JobModel, result: Any, job_info_ttl: int, result_ttl: int, connection: ConnectionType
    ):
        """Saves and cleanup job after successful execution"""
        job.after_execution(
            job_info_ttl,
            JobStatus.FINISHED,
            prev_registry=self.active_job_registry,
            new_registry=self.finished_job_registry,
            connection=connection,
        )
        Result.create(
            connection,
            job_name=job.name,
            worker_name=job.worker_name,
            _type=ResultType.SUCCESSFUL,
            return_value=result,
            ttl=result_ttl,
        )

    def job_handle_failure(self, status: JobStatus, job: JobModel, exc_string: str, connection: ConnectionType):
        # Does not set job status since the job might be stopped
        job.after_execution(
            SCHEDULER_CONFIG.DEFAULT_FAILURE_TTL,
            status,
            prev_registry=self.active_job_registry,
            new_registry=self.failed_job_registry,
            connection=connection,
        )
        Result.create(
            connection,
            job.name,
            job.worker_name,
            ResultType.FAILED,
            SCHEDULER_CONFIG.DEFAULT_FAILURE_TTL,
            exc_string=exc_string,
        )

    def run_sync(self, job: JobModel) -> JobModel:
        """Run a job synchronously, meaning on the same process the method was called."""
        job.prepare_for_execution("sync", self.active_job_registry, self.connection)
        try:
            result = perform_job(job, self.connection)

            with self.connection.pipeline() as pipeline:
                self.job_handle_success(
                    job, result=result, job_info_ttl=job.job_info_ttl, result_ttl=job.success_ttl, connection=pipeline
                )

                pipeline.execute()
        except Exception as e:  # noqa
            logger.warning(f"Job {job.name} failed with exception: {e}")
            with self.connection.pipeline() as pipeline:
                exc_string = "".join(traceback.format_exception(*sys.exc_info()))
                self.job_handle_failure(JobStatus.FAILED, job, exc_string, pipeline)
                pipeline.execute()
        return job

    @classmethod
    def dequeue_any(
        cls,
        queues: List[Self],
        timeout: Optional[int],
        connection: Optional[ConnectionType] = None,
    ) -> Tuple[Optional[JobModel], Optional[Self]]:
        """Class method returning a Job instance at the front of the given set of Queues, where the order of the queues
        is important.

        When all the Queues are empty, depending on the `timeout` argument, either blocks execution of this function
        for the duration of the timeout or until new messages arrive on any of the queues, or returns None.

        :param queues: List of Queue objects
        :param timeout: Timeout for the pop operation
        :param connection: Broker Connection
        :returns: Tuple of Job, Queue
        """

        while True:
            registries = [q.queued_job_registry for q in queues]
            for registry in registries:
                registry.compact()

            registry_key, job_name = QueuedJobRegistry.pop(connection, registries, timeout)
            if job_name is None:
                return None, None

            queue = next(filter(lambda q: q.queued_job_registry.key == registry_key, queues), None)
            if queue is None:
                logger.warning(f"Could not find queue for registry key {registry_key} in queues")
                return None, None

            job = JobModel.get(job_name, connection=connection)
            if job is None:
                continue
            return job, queue
        return None, None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name!r})"

    def __str__(self) -> str:
        return f"<{self.__class__.__name__} {self.name}>"

    def _remove_from_registries(self, job_name: str, connection: ConnectionType) -> None:
        """Removes the job from all registries besides failed_job_registry"""
        self.finished_job_registry.delete(connection=connection, job_name=job_name)
        self.scheduled_job_registry.delete(connection=connection, job_name=job_name)
        self.active_job_registry.delete(connection=connection, job_name=job_name)
        self.canceled_job_registry.delete(connection=connection, job_name=job_name)
        self.queued_job_registry.delete(connection=connection, job_name=job_name)

    def cancel_job(self, job_name: str) -> None:
        """Cancels the given job, which will prevent the job from ever running (or inspected).

        This method merely exists as a high-level API call to cancel jobs without worrying about the internals required
        to implement job cancellation.

        :param job_name: The job name to cancel.
        :raises NoSuchJobError: If the job does not exist.
        :raises InvalidJobOperation: If the job has already been canceled.
        """
        job = JobModel.get(job_name, connection=self.connection)
        if job is None:
            raise NoSuchJobError(f"No such job: {job_name}")
        if job.status == JobStatus.CANCELED:
            raise InvalidJobOperation(f"Cannot cancel already canceled job: {job.name}")

        pipe = self.connection.pipeline()
        new_status = JobStatus.CANCELED if job.status == JobStatus.QUEUED else JobStatus.STOPPED

        while True:
            try:
                job.set_field("status", new_status, connection=pipe)
                self._remove_from_registries(job_name, connection=pipe)
                pipe.execute()
                if new_status == JobStatus.CANCELED:
                    self.canceled_job_registry.add(pipe, job_name, 0)
                else:
                    self.finished_job_registry.add(
                        pipe, job_name, current_timestamp() + SCHEDULER_CONFIG.DEFAULT_FAILURE_TTL
                    )
                pipe.execute()
                break
            except WatchError:
                # if the pipeline comes from the caller, we re-raise the exception as it is the responsibility of the
                # caller to handle it
                raise

    def delete_job(self, job_name: str, expire_job_model: bool = True) -> None:
        """Deletes the given job from the queue and all its registries"""
        pipe = self.connection.pipeline()

        while True:
            try:
                self._remove_from_registries(job_name, connection=pipe)
                self.failed_job_registry.delete(connection=pipe, job_name=job_name)
                if expire_job_model:
                    job_model = JobModel.get(job_name, connection=self.connection)
                    if job_model is not None:
                        job_model.expire(ttl=job_model.job_info_ttl, connection=pipe)
                pipe.execute()
                break
            except WatchError:
                pass

    def enqueue_job(
        self, job_model: JobModel, connection: Optional[ConnectionType] = None, at_front: bool = False
    ) -> JobModel:
        """Enqueues a job for delayed execution without checking dependencies.

        If Queue is instantiated with is_async=False, job is executed immediately.
        :param job_model: The job redis model
        :param connection: The Redis Pipeline
        :param at_front: Whether to enqueue the job at the front

        :returns: The enqueued JobModel
        """

        pipe = connection if connection is not None else self.connection.pipeline()
        job_model.started_at = None
        job_model.ended_at = None
        job_model.status = JobStatus.QUEUED
        job_model.enqueued_at = utcnow()
        job_model.save(connection=pipe)

        if self._is_async:
            if at_front:
                score = current_timestamp()
            else:
                score = self.queued_job_registry.get_last_timestamp() or current_timestamp()
            self.scheduled_job_registry.delete(connection=pipe, job_name=job_model.name)
            self.queued_job_registry.add(connection=pipe, score=score, job_name=job_model.name)
            pipe.execute()
            logger.debug(f"Pushed job {job_model.name} into {self.name} queued-jobs registry")
        else:  # sync mode
            pipe.execute()
            job_model = self.run_sync(job_model)
            job_model.expire(ttl=job_model.job_info_ttl, connection=pipe)
            pipe.execute()

        return job_model



# Generated by Django 5.1.7 on 2025-03-24 14:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('scheduler', '0020_remove_repeatabletask_new_task_id_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='task',
            name='job_id',
        ),
        migrations.AddField(
            model_name='task',
            name='job_name',
            field=models.CharField(blank=True, editable=False, help_text='Current job_name on queue', max_length=128, null=True, verbose_name='job name'),
        ),
    ]



import json
import os
import tempfile
from unittest import mock

import yaml
from django.core.management import call_command
from django.test import TestCase

from scheduler.tests import test_settings  # noqa
from scheduler.tests.testtools import task_factory
from scheduler.models import TaskType


class ExportTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.tmpfile = tempfile.NamedTemporaryFile()

    def tearDown(self) -> None:
        super().tearDown()
        os.remove(self.tmpfile.name)

    def test_export__should_export_job(self):
        tasks = list()
        tasks.append(task_factory(TaskType.ONCE, enabled=True))
        tasks.append(task_factory(TaskType.REPEATABLE, enabled=True))

        # act
        call_command("export", filename=self.tmpfile.name)
        # assert
        result = json.load(self.tmpfile)
        self.assertEqual(len(tasks), len(result))
        self.assertEqual(result[0], tasks[0].to_dict())
        self.assertEqual(result[1], tasks[1].to_dict())

    def test_export__should_export_enabled_jobs_only(self):
        tasks = list()
        tasks.append(task_factory(TaskType.ONCE, enabled=True))
        tasks.append(task_factory(TaskType.REPEATABLE, enabled=False))

        # act
        call_command("export", filename=self.tmpfile.name, enabled=True)
        # assert
        result = json.load(self.tmpfile)
        self.assertEqual(len(tasks) - 1, len(result))
        self.assertEqual(result[0], tasks[0].to_dict())

    def test_export__should_export_job_yaml_without_yaml_lib(self):
        tasks = list()
        tasks.append(task_factory(TaskType.ONCE, enabled=True))
        tasks.append(task_factory(TaskType.REPEATABLE, enabled=True))

        # act
        with mock.patch.dict("sys.modules", {"yaml": None}):
            with self.assertRaises(SystemExit) as cm:
                call_command("export", filename=self.tmpfile.name, format="yaml")
            self.assertEqual(cm.exception.code, 1)

    def test_export__should_export_job_yaml_green(self):
        tasks = list()
        tasks.append(task_factory(TaskType.ONCE, enabled=True))
        tasks.append(task_factory(TaskType.REPEATABLE, enabled=True))
        tasks.append(task_factory(TaskType.CRON, enabled=True))

        # act
        call_command("export", filename=self.tmpfile.name, format="yaml")
        # assert
        result = yaml.load(self.tmpfile, yaml.SafeLoader)
        self.assertEqual(len(tasks), len(result))
        self.assertEqual(result[0], tasks[0].to_dict())
        self.assertEqual(result[1], tasks[1].to_dict())
        self.assertEqual(result[2], tasks[2].to_dict())



import time
from datetime import datetime

from django.urls import reverse

from scheduler.helpers.queues import get_queue
from scheduler.tests.jobs import test_job
from scheduler.tests.test_views.base import BaseTestCase


class QueueRegistryJobsViewTest(BaseTestCase):
    def test_queue_jobs_unknown_registry(self):
        queue_name = "default"
        res = self.client.get(reverse("queue_registry_jobs", args=[queue_name, "unknown"]), follow=True)
        self.assertEqual(404, res.status_code)

    def test_queue_jobs_unknown_queue(self):
        res = self.client.get(reverse("queue_registry_jobs", args=["UNKNOWN", "queued"]))
        self.assertEqual(404, res.status_code)

    def test_queued_jobs(self):
        """Jobs in queue are displayed properly"""
        queue = get_queue("default")
        job = queue.create_and_enqueue_job(test_job)
        queue_name = "default"
        res = self.client.get(reverse("queue_registry_jobs", args=[queue_name, "queued"]))
        self.assertEqual(res.context["jobs"], [job])

    def test_finished_jobs(self):
        """Ensure that finished jobs page works properly."""
        queue = get_queue("django_tasks_scheduler_test")
        queue_name = "django_tasks_scheduler_test"

        job = queue.create_and_enqueue_job(test_job)
        registry = queue.finished_job_registry
        registry.add(queue.connection, job.name, time.time() + 2)
        res = self.client.get(reverse("queue_registry_jobs", args=[queue_name, "finished"]))
        self.assertEqual(res.context["jobs"], [job])

    def test_failed_jobs(self):
        """Ensure that failed jobs page works properly."""
        queue = get_queue("django_tasks_scheduler_test")
        queue_name = "django_tasks_scheduler_test"

        # Test that page doesn't fail when FailedJobRegistry is empty
        res = self.client.get(reverse("queue_registry_jobs", args=[queue_name, "failed"]))
        self.assertEqual(res.status_code, 200)

        job = queue.create_and_enqueue_job(test_job)
        registry = queue.failed_job_registry
        registry.add(queue.connection, job.name, time.time() + 20)
        res = self.client.get(reverse("queue_registry_jobs", args=[queue_name, "failed"]))
        self.assertEqual(res.context["jobs"], [job])

    def test_scheduled_jobs(self):
        """Ensure that scheduled jobs page works properly."""
        queue = get_queue("django_tasks_scheduler_test")
        queue_name = "django_tasks_scheduler_test"

        # Test that page doesn't fail when ScheduledJobRegistry is empty
        res = self.client.get(reverse("queue_registry_jobs", args=[queue_name, "scheduled"]))
        self.assertEqual(res.status_code, 200)

        job = queue.create_and_enqueue_job(test_job, when=datetime.now())
        res = self.client.get(reverse("queue_registry_jobs", args=[queue_name, "scheduled"]))
        self.assertEqual(res.context["jobs"], [job])

    def test_scheduled_jobs_registry_removal(self):
        """Ensure that non-existing job is being deleted from registry by view"""
        queue = get_queue("django_tasks_scheduler_test")
        queue_name = "django_tasks_scheduler_test"

        registry = queue.scheduled_job_registry
        job = queue.create_and_enqueue_job(test_job, when=datetime.now())
        self.assertEqual(len(registry), 1)

        queue.delete_job(job.name)
        res = self.client.get(reverse("queue_registry_jobs", args=[queue_name, "scheduled"]))
        self.assertEqual(res.context["jobs"], [])

        self.assertEqual(len(registry), 0)

    def test_started_jobs(self):
        """Ensure that active jobs page works properly."""
        queue = get_queue("django_tasks_scheduler_test")
        queue_name = "django_tasks_scheduler_test"

        job = queue.create_and_enqueue_job(test_job)
        registry = queue.active_job_registry
        registry.add(queue.connection, job.name, time.time() + 20)
        res = self.client.get(reverse("queue_registry_jobs", args=[queue_name, "active"]))
        self.assertEqual(res.context["jobs"], [job])



import os
import signal

from scheduler.types import ConnectionType
from scheduler.settings import logger
from scheduler.worker.commands.worker_commands import WorkerCommand


class ShutdownCommand(WorkerCommand):
    """shutdown command"""

    command_name = "shutdown"

    def process_command(self, connection: ConnectionType) -> None:
        logger.info("Received shutdown command, sending SIGINT signal.")
        pid = os.getpid()
        os.kill(pid, signal.SIGINT)



"""list_registry_jobs actions on multiple selected jobs"""

from enum import Enum

from django.contrib import admin, messages
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse, HttpRequest
from django.shortcuts import render, redirect
from django.urls import reverse
from django.views.decorators.cache import never_cache

from scheduler.redis_models import JobModel
from scheduler.settings import logger
from scheduler.views.helpers import get_queue, _check_next_url, _enqueue_multiple_jobs
from scheduler.worker.commands import StopJobCommand, send_command


class QueueJobAction(Enum):
    DELETE = "delete"
    REQUEUE = "requeue"
    STOP = "stop"

    def __contains__(self, item) -> bool:
        return item in [a.value for a in self.__class__]


@never_cache
@staff_member_required
def queue_job_actions(request: HttpRequest, queue_name: str) -> HttpResponse:
    queue = get_queue(queue_name)
    next_url = _check_next_url(request, reverse("queue_registry_jobs", args=[queue_name, "queued"]))
    action = request.POST.get("action", False)
    job_names = request.POST.get("job_names", False)
    if request.method != "POST" or not action or not job_names or action not in [item.value for item in QueueJobAction]:
        return redirect(next_url)
    job_names = request.POST.getlist("job_names")
    if action == QueueJobAction.DELETE.value:
        jobs = JobModel.get_many(job_names, connection=queue.connection)
        for job in jobs:
            if job is None:
                continue
            queue.delete_job(job.name)
        messages.info(request, f"You have successfully deleted {len(job_names)} jobs!")
    elif action == QueueJobAction.REQUEUE.value:
        requeued_jobs_count = _enqueue_multiple_jobs(queue, job_names)
        messages.info(request, f"You have successfully re-queued {requeued_jobs_count}/{len(job_names)}  jobs!")
    elif action == QueueJobAction.STOP.value:
        cancelled_jobs = 0
        jobs = JobModel.get_many(job_names, connection=queue.connection)
        for job in jobs:
            if job is None:
                continue
            try:
                command = StopJobCommand(job_name=job.name, worker_name=job.worker_name)
                send_command(connection=queue.connection, command=command)
                queue.cancel_job(job.name)
                cancelled_jobs += 1
            except Exception as e:
                logger.warning(f"Could not stop job: {e}")
                pass
        messages.info(request, f"You have successfully stopped {cancelled_jobs}  jobs!")
    return redirect(next_url)


@never_cache
@staff_member_required
def queue_confirm_job_action(request: HttpRequest, queue_name: str) -> HttpResponse:
    queue = get_queue(queue_name)
    next_url = _check_next_url(request, reverse("queue_registry_jobs", args=[queue_name, "queued"]))
    action = request.POST.get("action", None)
    job_names = request.POST.getlist("_selected_action", None)
    if request.method != "POST" or action is None or job_names is None or action not in QueueJobAction:
        return redirect(next_url)

    # confirm action
    context_data = {
        **admin.site.each_context(request),
        "action": action,
        "jobs": [JobModel.get(job_name, connection=queue.connection) for job_name in job_names],
        "total_jobs": len(job_names),
        "queue": queue,
        "next_url": next_url,
        "action_url": reverse(
            "queue_job_actions",
            args=[
                queue_name,
            ],
        ),
    }
    return render(request, "admin/scheduler/confirm_action.html", context_data)



import time
from datetime import datetime, timedelta, timezone
from typing import ClassVar, Optional, List, Tuple

from scheduler.helpers.utils import current_timestamp
from scheduler.types import ConnectionType
from .base_registry import JobNamesRegistry
from .. import as_str
from ..job import JobModel


class QueuedJobRegistry(JobNamesRegistry):
    _element_key_template: ClassVar[str] = ":registry:{}:queued_jobs"

    def cleanup(self, connection: ConnectionType, timestamp: Optional[float] = None) -> None:
        """This method is only here to prevent errors because this method is automatically called by `count()`
        and `all()` methods implemented in JobIdsRegistry."""
        pass

    def compact(self) -> None:
        """Removes all "dead" jobs from the queue by cycling through it, while guaranteeing FIFO semantics."""
        jobs_with_ts = self.all_with_timestamps()
        for job_name, timestamp in jobs_with_ts:
            if not JobModel.exists(job_name, self.connection):
                self.delete(connection=self.connection, job_name=job_name)

    def empty(self) -> None:
        queued_jobs_count = self.count(connection=self.connection)
        with self.connection.pipeline() as pipe:
            for offset in range(0, queued_jobs_count, 1000):
                job_names = self.all(offset, 1000)
                for job_name in job_names:
                    self.delete(connection=pipe, job_name=job_name)
                JobModel.delete_many(job_names, connection=pipe)
            pipe.execute()


class FinishedJobRegistry(JobNamesRegistry):
    _element_key_template: ClassVar[str] = ":registry:{}:finished_jobs"


class FailedJobRegistry(JobNamesRegistry):
    _element_key_template: ClassVar[str] = ":registry:{}:failed_jobs"


class CanceledJobRegistry(JobNamesRegistry):
    _element_key_template: ClassVar[str] = ":registry:{}:canceled_jobs"

    def cleanup(self, connection: ConnectionType, timestamp: Optional[float] = None) -> None:
        """This method is only here to prevent errors because this method is automatically called by `count()`
        and `all()` methods implemented in JobIdsRegistry."""
        pass


class ScheduledJobRegistry(JobNamesRegistry):
    _element_key_template: ClassVar[str] = ":registry:{}:scheduled_jobs"

    def cleanup(self, connection: ConnectionType, timestamp: Optional[float] = None) -> None:
        """This method is only here to prevent errors because this method is automatically called by `count()`
        and `all()` methods implemented in JobIdsRegistry."""
        pass

    def schedule(self, connection: ConnectionType, job_name: str, scheduled_datetime: datetime) -> int:
        """Adds job_name to registry, scored by its execution time (in UTC).
        If datetime has no tzinfo, it will assume localtimezone.

        :param connection: Broker connection
        :param job_name: Job name to schedule
        :param scheduled_datetime: datetime to schedule job
        """
        # If datetime has no timezone, assume server's local timezone
        if not scheduled_datetime.tzinfo:
            tz = timezone(timedelta(seconds=-(time.timezone if time.daylight == 0 else time.altzone)))
            scheduled_datetime = scheduled_datetime.replace(tzinfo=tz)

        timestamp = scheduled_datetime.timestamp()
        return self.add(connection=connection, job_name=job_name, score=timestamp)

    def get_jobs_to_schedule(self, timestamp: int, chunk_size: int = 1000) -> List[str]:
        """Gets a list of job names that should be scheduled.

        :param timestamp: timestamp/score of jobs in SortedSet.
        :param chunk_size: Max results to return.
        :returns: A list of job names
        """
        jobs_to_schedule = self.connection.zrangebyscore(self._key, 0, max=timestamp, start=0, num=chunk_size)
        return [as_str(job_name) for job_name in jobs_to_schedule]

    def get_scheduled_time(self, job_name: str) -> Optional[datetime]:
        """Returns datetime (UTC) at which job is scheduled to be enqueued

        :param job_name: Job name
        :returns: The scheduled time as datetime object, or None if job is not found
        """

        score: Optional[float] = self.connection.zscore(self._key, job_name)
        if not score:
            return None

        return datetime.fromtimestamp(score, tz=timezone.utc)


class ActiveJobRegistry(JobNamesRegistry):
    """Registry of currently executing jobs. Each queue maintains a ActiveJobRegistry."""

    _element_key_template: ClassVar[str] = ":registry:{}:active"

    def get_job_names_before(self, connection: ConnectionType, timestamp: Optional[float]) -> List[Tuple[str, float]]:
        """Returns job names whose score is lower than a timestamp timestamp.

        Returns names for jobs with an expiry time earlier than timestamp,
        specified as seconds since the Unix epoch.
        timestamp defaults to calltime if unspecified.
        """
        score = timestamp or current_timestamp()
        jobs_before = connection.zrangebyscore(self._key, 0, score, withscores=True)
        return [(as_str(job_name), score) for (job_name, score) in jobs_before]



from scheduler.helpers.queues import get_queue
from scheduler.tests.jobs import test_job
from ..test_views.base import BaseTestCase
from ...redis_models import JobModel
from ...worker import create_worker
from ...worker.commands import send_command
from ...worker.commands.suspend_worker import SuspendWorkCommand


class WorkerCommandsTest(BaseTestCase):
    def test_stop_worker_command__green(self):
        # Arrange
        worker_name = "test"
        queue = get_queue("default")
        job = queue.create_and_enqueue_job(test_job)
        self.assertTrue(job.is_queued)
        worker = create_worker("default", name=worker_name, burst=True, with_scheduler=False)
        worker.worker_start()
        # Act
        send_command(queue.connection, SuspendWorkCommand(worker_name=worker_name))
        worker.work()

        # Assert
        self.assertTrue(job.is_queued)
        self.assertTrue(worker._model.is_suspended)

    def test_stop_worker_command__bad_worker_name(self):
        # Arrange
        worker_name = "test"
        queue = get_queue("default")
        job = queue.create_and_enqueue_job(test_job)
        self.assertTrue(job.is_queued)
        worker = create_worker("default", name=worker_name, burst=True, with_scheduler=False)
        worker.bootstrap()
        # Act
        send_command(queue.connection, SuspendWorkCommand(worker_name=worker_name + "1"))
        worker.work()

        # Assert
        self.assertFalse(worker._model.is_suspended)
        job = JobModel.get(job.name, connection=queue.connection)
        self.assertFalse(job.is_queued)



from time import sleep

from django.test import tag

from scheduler.helpers.queues import get_queue
from scheduler.redis_models import JobStatus, JobModel, WorkerModel
from scheduler.tests.jobs import long_job, two_seconds_job
from .. import testtools
from ..test_views.base import BaseTestCase
from ...worker.commands import KillWorkerCommand, send_command, StopJobCommand


@tag("multiprocess")
class WorkerCommandsTest(BaseTestCase):
    def test_kill_job_command__current_job(self):
        # Arrange
        queue = get_queue("django_tasks_scheduler_test")
        job = queue.create_and_enqueue_job(long_job)
        self.assertTrue(job.is_queued)
        process, worker_name = testtools.run_worker_in_process("django_tasks_scheduler_test")
        sleep(0.1)
        job = JobModel.get(job.name, connection=queue.connection)
        self.assertEqual(JobStatus.STARTED, job.status)

        # Act
        send_command(queue.connection, StopJobCommand(worker_name=worker_name, job_name=job.name))

        # Assert

        process.terminate()
        process.join(2)
        process.kill()

        job = JobModel.get(job.name, connection=queue.connection)
        worker_model = WorkerModel.get(worker_name, connection=queue.connection)
        self.assertEqual(job.name, worker_model.stopped_job_name)
        self.assertEqual(job.name, worker_model.current_job_name)
        self.assertEqual(0, worker_model.completed_jobs)
        self.assertEqual(0, worker_model.failed_job_count)
        self.assertEqual(0, worker_model.successful_job_count)
        self.assertEqual(JobStatus.STOPPED, job.status)
        self.assertNotIn(job.name, queue.queued_job_registry.all())

    def test_kill_job_command__different_job(self):
        # Arrange
        queue = get_queue("django_tasks_scheduler_test")
        job = queue.create_and_enqueue_job(two_seconds_job)
        self.assertTrue(job.is_queued)
        process, worker_name = testtools.run_worker_in_process("django_tasks_scheduler_test")
        sleep(0.2)
        job = JobModel.get(job.name, connection=queue.connection)
        self.assertEqual(JobStatus.STARTED, job.status)

        # Act
        send_command(queue.connection, StopJobCommand(worker_name=worker_name, job_name=job.name + "1"))
        sleep(0.1)
        process.kill()
        process.join()
        # Assert
        job = JobModel.get(job.name, connection=queue.connection)
        self.assertEqual(JobStatus.STARTED, job.status)
        self.assertNotIn(job.name, queue.queued_job_registry.all())
        worker_model = WorkerModel.get(worker_name, connection=queue.connection)
        self.assertEqual(0, worker_model.completed_jobs)
        self.assertEqual(0, worker_model.failed_job_count)
        self.assertEqual(0, worker_model.successful_job_count)
        self.assertIsNone(worker_model.stopped_job_name)
        self.assertEqual(job.name, worker_model.current_job_name)

    def test_kill_worker_command(self):
        queue = get_queue("django_tasks_scheduler_test")
        process, worker_name = testtools.run_worker_in_process("django_tasks_scheduler_test")
        sleep(0.1)
        # act
        send_command(queue.connection, KillWorkerCommand(worker_name=worker_name))
        # assert
        sleep(0.2)
        process.kill()
        process.join()
        worker_model = WorkerModel.get(worker_name, connection=queue.connection)
        self.assertEqual(0, worker_model.completed_jobs)
        self.assertEqual(0, worker_model.failed_job_count)
        self.assertEqual(0, worker_model.successful_job_count)
        self.assertIsNotNone(worker_model.shutdown_requested_date)



from datetime import timedelta, datetime

from django.core.exceptions import ValidationError
from django.utils import timezone

from scheduler import settings
from scheduler.models import TaskType
from scheduler.tests.test_task_types.test_task_model import BaseTestCases
from scheduler.tests.testtools import task_factory


class TestScheduledTask(BaseTestCases.TestSchedulableTask):
    task_type = TaskType.ONCE
    queue_name = settings.get_queue_names()[0]

    def test_clean(self):
        job = task_factory(self.task_type)
        job.queue = self.queue_name
        job.callable = "scheduler.tests.jobs.test_job"
        self.assertIsNone(job.clean())

    def test_create_without_date__fail(self):
        task = task_factory(self.task_type, scheduled_time=None, instance_only=True)
        self.assertIsNone(task.scheduled_time)
        with self.assertRaises(Exception) as cm:
            task.clean()
        self.assertTrue(isinstance(cm.exception, ValidationError))
        self.assertEqual(str(cm.exception), "{'scheduled_time': ['Scheduled time is required']}")

    def test_create_with_date_in_the_past__fail(self):
        task = task_factory(self.task_type, scheduled_time=datetime.now() - timedelta(days=1), instance_only=True)
        with self.assertRaises(Exception) as cm:
            task.clean()
        self.assertTrue(isinstance(cm.exception, ValidationError))
        self.assertEqual(str(cm.exception), "{'scheduled_time': ['Scheduled time must be in the future']}")

    def test_unschedulable_old_job(self):
        job = task_factory(self.task_type, scheduled_time=timezone.now() - timedelta(hours=1))
        self.assertFalse(job.is_scheduled())
